# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import time
from typing import Any, Optional, Set, Tuple

from supervisor.loggers import Logger

from .context import Context
from .instancestatus import SupvisorsInstanceStatus
from .options import SupvisorsOptions
from .process import ProcessStatus
from .strategy import conciliate_conflicts
from .ttypes import (SupvisorsInstanceStates, RunningFailureStrategies, SupvisorsStates, SynchronizationOptions,
                     NameList, Payload, PayloadList, WORKING_STATES)


class Forced:
    pass


class AbstractState:
    """ Base class for a state with simple entry / next / exit actions.

    Attributes are:
        - supvisors: the reference to the global Supvisors structure ;
        - context: the reference to the context of the global Supvisors structure ;
        - logger: the reference to the logger of the global Supvisors structure ;
        - local_identifier: the identifier of the local Supvisors instance.
     """

    def __init__(self, supvisors: Any) -> None:
        """ Initialization of the attributes.

        :param supvisors: the global Supvisors structure
        """
        self.supvisors = supvisors
        self.local_identifier = supvisors.mapper.local_identifier

    @property
    def context(self) -> Context:
        """ Shortcut to the Supvisors context structure

        :return:
        """
        return self.supvisors.context

    @property
    def logger(self) -> Logger:
        """ Shortcut to the Supvisors logger structure

        :return:
        """
        return self.supvisors.logger

    def enter(self) -> None:
        """ Actions performed when entering the state.
        May be redefined in subclasses.

        :return: None
        """

    def next(self) -> SupvisorsStates:
        """ Actions performed upon reception of an event.
        May be redefined in subclasses.

        :return: None
        """

    def exit(self) -> None:
        """ Actions performed when leaving the state.
        May be redefined in subclasses.

        :return: None
        """

    def check_instances(self) -> Optional[SupvisorsStates]:
        """ Check that local and Master Supvisors instances are still RUNNING.
        If their ticks are not received anymore, back to INITIALIZATION state to force a synchronization phase.

        Set CHECKED Supvisors instances to RUNNING if no start sequence is in progress (in order to avoid interference).

        :return: the suggested state if local or Master Supvisors instance is not active anymore
        """
        # set all CHECKED Supvisors instances to RUNNING if no starting sequence is in progress
        if not self.supvisors.starter.in_progress():
            self.context.activate_checked()
        # check that the local Supvisors instance is still RUNNING
        if self.context.local_status.state != SupvisorsInstanceStates.RUNNING:
            self.logger.critical('AbstractState.check_instances: local Supvisors instance not RUNNING')
            return SupvisorsStates.INITIALIZATION
        # check that the Master Supvisors instance is still RUNNING
        master_instance = self.context.master_instance
        if not master_instance or master_instance.state != SupvisorsInstanceStates.RUNNING:
            self.logger.warn('AbstractState.check_instances: no Master Supvisors instance RUNNING')
            # in INITIALIZATION, master_identifier is not reset (state is not re-entered)
            self.context.master_identifier = ''
            return SupvisorsStates.INITIALIZATION
        return None

    def abort_jobs(self) -> None:
        """ Abort starting jobs in progress.

        :return: None
        """
        self.supvisors.failure_handler.abort()
        self.supvisors.starter.abort()
        self.supvisors.stopper.abort()


class OffState(AbstractState):

    def next(self) -> SupvisorsStates:
        """ Wait for Supervisor to be RUNNING.

        :return: the new Supvisors state
        """
        # The Supvisors XML-RPC handler is an easy mark to know that Supervisor is running
        if self.supvisors.rpc_handler:
            return SupvisorsStates.INITIALIZATION
        return SupvisorsStates.OFF


class InitializationState(AbstractState):
    """ In the INITIALIZATION state, Supvisors synchronizes to all known Supvisors instances. """

    def enter(self) -> None:
        """ When entering the INITIALIZATION state, reset the context.

        :return: None
        """
        # clear any existing job
        self.abort_jobs()
        # reset context, keeping the isolation status
        self.context.reset()

    def _check_end_sync_strict(self) -> bool:
        """ End of sync phase if all Supvisors instances declared in the supvisors_list option are running.
        NOTE: this option is NOT allowed if the supvisors_list is empty (which is expected in discovery mode,
        but still possible).
        """
        if SynchronizationOptions.STRICT in self.supvisors.options.synchro_options:
            if self.context.initial_running():
                self.logger.info('InitializationState._check_end_sync_list: all expected Supvisors instances'
                                 ' are RUNNING')
                return True
        return False

    def _check_end_sync_list(self) -> bool:
        """ End of sync phase if all known Supvisors instances are running.
        NOTE: this includes the Supvisors instances declared in the supvisors_list option and the discovered
        Supvisors instances.
        """
        if SynchronizationOptions.LIST in self.supvisors.options.synchro_options:
            if self.context.all_running():
                self.logger.info('InitializationState._check_end_sync_list: all known Supvisors instances'
                                 ' are RUNNING')
                return True
        return False

    def _check_end_sync_timeout(self, uptime: float) -> bool:
        """ End of sync phase if the uptime has exceeded the synchro_timeout. """
        if SynchronizationOptions.TIMEOUT in self.supvisors.options.synchro_options:
            synchro_timeout = self.supvisors.options.synchro_timeout
            self.logger.debug(f'InitializationState._check_end_sync_timeout: uptime={uptime}'
                              f' synchro_timeout={synchro_timeout}')
            if uptime >= synchro_timeout:
                self.logger.info(f'InitializationState._check_end_sync_timeout: timeout {synchro_timeout} reached')
                return True
        return False

    def _check_end_sync_core(self, uptime: float, running_identifiers: NameList) -> bool:
        """ End of sync phase if all core Supvisors instances are in a known state.
        If the condition is reached, the DEPLOYMENT state may be reached with Supvisors instances still UNKNOWN.
        NOTE: this option is NOT allowed if the core_identifiers is empty (which is expected in discovery mode,
        but still possible). """
        if SynchronizationOptions.CORE in self.supvisors.options.synchro_options:
            # all core Supvisors instances must be running
            # in case of late start, a security limit of SYNCHRO_TIMEOUT_MIN is kept to give a chance
            # to other Supvisors instances and limit the number of redeploy_mark
            if uptime > SupvisorsOptions.SYNCHRO_TIMEOUT_MIN and self.context.running_core_identifiers():
                # in the event where the local Supvisors instance has been started lately, there may be already
                #   a Master Supvisors instance, in which case it must be seen as running
                if not self.context.master_identifier or self.context.master_identifier in running_identifiers:
                    self.logger.info('InitializationState._check_end_sync_core: all core Supvisors instances'
                                     ' are RUNNING')
                    return True
                self.logger.info('InitializationState._check_end_sync_core: all core Supvisors instances'
                                 f' are RUNNING but not the declared Master={self.context.master_identifier}')
        return False

    def _check_end_sync_user(self, running_identifiers: NameList) -> bool:
        """ End of sync phase if the master is known.
        This is meant to be triggered using the Web UI or using the XML-RPC API.
        No time condition applies as the user is responsible.
        """
        if SynchronizationOptions.USER in self.supvisors.options.synchro_options:
            # the Master Supvisors instance must be seen as running
            if self.context.master_identifier and self.context.master_identifier in running_identifiers:
                self.logger.info('InitializationState._check_end_sync_user: the Supvisors Master instance is RUNNING')
                return True
        return False

    def next(self) -> SupvisorsStates:
        """ Wait for Supvisors instances to exchange their data until a condition is reached
        to end the synchronization phase.

        :return: the new Supvisors state
        """
        # set all CHECKED Supvisors instances to RUNNING
        self.context.activate_checked()
        # get duration from start date
        uptime: float = time.monotonic() - self.context.start_date
        # cannot get out of this state without local Supvisors instance RUNNING
        running_identifiers = self.context.running_identifiers()
        if self.local_identifier in running_identifiers:
            # check end of sync conditions
            self.logger.trace(f'InitializationState.next: synchro_options={self.supvisors.options.synchro_options}')
            strict_sync = self._check_end_sync_strict()
            list_sync = self._check_end_sync_list()
            timeout_synch = self._check_end_sync_timeout(uptime)
            core_sync = self._check_end_sync_core(uptime, running_identifiers)
            user_sync = self._check_end_sync_user(running_identifiers)
            self.logger.debug(f'InitializationState.next: strict_sync={strict_sync} list_sync={list_sync}'
                              f' timeout_synch={timeout_synch} core_sync={core_sync} user_sync={user_sync}')
            if strict_sync or list_sync or timeout_synch or core_sync or user_sync:
                if self.context.master_identifier:
                    # check Master status and reset if not running
                    if self.context.master_instance.state != SupvisorsInstanceStates.RUNNING:
                        self.context.master_identifier = ''
                if not self.context.master_identifier:
                    self.context.elect_master()
                # The Master can exit the INITIALIZATION state by itself
                if self.context.is_master:
                    return SupvisorsStates.DISTRIBUTION
                # The Slaves will follow the Master state
                # WARN: at this point, the Master FSM state may not be known yet
                return self.supvisors.context.supvisors_state
        else:
            # log current status
            if uptime >= SupvisorsOptions.SYNCHRO_TIMEOUT_MIN:
                self.logger.critical(f'InitializationState.next: local Supvisors={self.local_identifier} still'
                                     f' not RUNNING after {int(uptime)} seconds')
            else:
                self.logger.debug(f'InitializationState.next: local Supvisors={self.local_identifier} still'
                                  f' not RUNNING after {int(uptime)} seconds')
        return SupvisorsStates.INITIALIZATION

    def exit(self):
        """ When exiting the INITIALIZATION state, set all non-responsive Supvisors instances to SILENT or ISOLATED.

        :return: None
        """
        self.context.invalid_unknown()
        # TODO 0.19: synchro_options are used only the first time
        #       in the event of a master loss or conflict, all instances are ready to communicate
        #       this will avoid 2 blocking states:
        #           1) STRICT only and one node becomes missing
        #           2) USER only and master lost (wait for user input again...)
        #           3) CORE only and one core node lost
        #       when it happens, applications are running and Supvisors is stuck in INIT, so impossible to act on apps
        #       introduce degraded mode for Supvisors ?
        #       on degraded: BLOCK, CONTINUE, STOP, SHUTDOWN


class MasterDeploymentState(AbstractState):
    """ In the DEPLOYMENT state, Supvisors starts automatically the applications having a starting model. """

    def enter(self):
        """ Trigger the automatic start and stop. """
        self.supvisors.starter.start_applications(self.supvisors.fsm.redeploy_mark is Forced)
        self.supvisors.fsm.redeploy_mark = False

    def next(self) -> SupvisorsStates:
        """ Check if the starting tasks are completed.

        :return: the new Supvisors state
        """
        # common check on local and Master Supvisors instances
        next_state = self.check_instances()
        if next_state:
            return next_state
        # Master goes to OPERATION when starting is completed
        if self.supvisors.starter.in_progress():
            return SupvisorsStates.DISTRIBUTION
        return SupvisorsStates.OPERATION


class MasterOperationState(AbstractState):
    """ In the OPERATION state, Supvisors is waiting for requests. """

    def next(self) -> SupvisorsStates:
        """ Check that all Supvisors instances are still active.
        Look after possible conflicts due to multiple running instances of the same process.

        :return: the new Supvisors state
        """
        # common check on local and Master Supvisors instances
        next_state = self.check_instances()
        if next_state:
            return next_state
        # check if jobs are in progress
        if self.supvisors.starter.in_progress() or self.supvisors.stopper.in_progress():
            return SupvisorsStates.OPERATION
        # check duplicated processes
        if self.context.conflicting():
            return SupvisorsStates.CONCILIATION
        # a redeployment mark has been set due to a new alive Supvisors instance
        # back to DEPLOYMENT state to repair what may have failed before
        if self.supvisors.fsm.redeploy_mark:
            return SupvisorsStates.DISTRIBUTION
        return SupvisorsStates.OPERATION


class MasterConciliationState(AbstractState):
    """ In the CONCILIATION state, Supvisors conciliates the conflicts. """

    def enter(self) -> None:
        """ When entering the CONCILIATION state, conciliate automatically the conflicts. """
        conciliate_conflicts(self.supvisors,
                             self.supvisors.options.conciliation_strategy,
                             self.context.conflicts())

    def next(self) -> SupvisorsStates:
        """ Check that all Supvisors instances are still active.
        Wait for all conflicts to be conciliated.

        :return: the new Supvisors state
        """
        # common check on local and Master Supvisors instances
        next_state = self.check_instances()
        if next_state:
            return next_state
        # check if jobs are in progress
        if self.supvisors.starter.in_progress() or self.supvisors.stopper.in_progress():
            return SupvisorsStates.CONCILIATION
        # back to OPERATION when there is no conflict anymore
        if not self.context.conflicting():
            return SupvisorsStates.OPERATION
        # new conflicts may happen while conciliation is in progress
        # call enter again to trigger a new conciliation
        self.enter()
        return SupvisorsStates.CONCILIATION


class MasterRestartingState(AbstractState):
    """ In the RESTARTING state, Supvisors stops all applications before triggering a full restart. """

    def enter(self) -> None:
        """ When entering the RESTARTING state, stop all applications.

        :return: None
        """
        self.abort_jobs()
        self.supvisors.stopper.stop_applications()

    def next(self) -> SupvisorsStates:
        """ Wait for all processes to be stopped.

        :return: the new Supvisors state
        """
        next_state = self.check_instances()
        if next_state:
            # no way going back to INITIALIZATION state at this point
            return SupvisorsStates.FINAL
        # check if stopping jobs are in progress
        if self.supvisors.stopper.in_progress():
            return SupvisorsStates.RESTARTING
        return SupvisorsStates.FINAL

    def exit(self):
        """ When exiting the RESTARTING state, request the Supervisor restart. """
        self.supvisors.rpc_handler.send_restart(self.local_identifier)
        # Slave instances will do the same when they will transition out of their RESTARTING state


class MasterShuttingDownState(AbstractState):
    """ In the SHUTTING_DOWN state, Supvisors stops all applications before triggering a full shutdown. """

    def enter(self):
        """ When entering in the SHUTTING_DOWN state, stop all applications. """
        self.abort_jobs()
        self.supvisors.stopper.stop_applications()

    def next(self):
        """ Wait for all processes to be stopped.

        :return: the new Supvisors state
        """
        # check eventual jobs in progress
        next_state = self.check_instances()
        if next_state:
            # no way going back to INITIALIZATION state at this point
            return SupvisorsStates.FINAL
        # check if stopping jobs are in progress
        if self.supvisors.stopper.in_progress():
            return SupvisorsStates.SHUTTING_DOWN
        return SupvisorsStates.FINAL

    def exit(self):
        """ When exiting the SHUTTING_DOWN state, request the Supervisor shutdown. """
        self.supvisors.rpc_handler.send_shutdown(self.local_identifier)
        # Slave instances will do the same when they will transition out of their SHUTTING_DOWN state


class SlaveMainState(AbstractState):

    def next(self) -> SupvisorsStates:
        """ The non-master instances are checking if local and Master Supvisors instances are still running.
        Local start / stop requests are evaluated too.

        :return: the new Supvisors state
        """
        # common check on local and Master Supvisors instances
        next_state = self.check_instances()
        if next_state:
            return next_state
        # next state is the Master state (maybe None)
        return self.supvisors.context.supvisors_state


class SlaveRestartingState(AbstractState):
    """ In the RESTARTING state, Supvisors stops all applications before triggering a full restart. """

    def enter(self) -> None:
        """ When entering the RESTARTING state, abort all pending tasks applications.

        :return: None
        """
        self.abort_jobs()

    def next(self) -> SupvisorsStates:
        """ Wait for all processes to be stopped.

        :return: the new Supvisors state
        """
        next_state = self.check_instances()
        if next_state:
            # no way going back to INITIALIZATION state at this point
            return SupvisorsStates.FINAL
        # stay in RESTARTING as long as the Master does
        if self.supvisors.context.supvisors_state == SupvisorsStates.RESTARTING:
            return SupvisorsStates.RESTARTING
        return SupvisorsStates.FINAL

    def exit(self):
        """ When exiting the RESTARTING state, request the full restart.

        NOTE: this has been moved from the former RestartState / enter because a Supvisors Slave
        could move from RESTARTING to any other state if the Master commands so,
        and it's important that Supervisor restarts at this point. """
        self.supvisors.rpc_handler.send_restart(self.local_identifier)


class SlaveShuttingDownState(AbstractState):
    """ In the SHUTTING_DOWN state, Supvisors stops all applications before triggering a full shutdown. """

    def enter(self) -> None:
        """ When entering the SHUTTING_DOWN state, abort all pending tasks applications.

        :return: None
        """
        self.abort_jobs()

    def next(self) -> SupvisorsStates:
        """ Wait for all processes to be stopped.

        :return: the new Supvisors state
        """
        next_state = self.check_instances()
        if next_state:
            # no way going back to INITIALIZATION state at this point
            return SupvisorsStates.FINAL
        # stay in SHUTTING_DOWN as long as the Master does
        if self.supvisors.context.supvisors_state == SupvisorsStates.SHUTTING_DOWN:
            return SupvisorsStates.SHUTTING_DOWN
        return SupvisorsStates.FINAL

    def exit(self):
        """ When exiting the SHUTTING_DOWN state, request the Supervisor shutdown.

        NOTE: this has been moved from the former ShutDownState / enter because a Supvisors Slave
        could move from SHUTTING_DOWN to any other state if the Master commands so,
        and it's important that Supervisor shuts down at this point. """
        self.supvisors.rpc_handler.send_shutdown(self.local_identifier)


class FinalState(AbstractState):
    """ This is a final state for Master and Slaves.
    Whatever it is a shutdown or a restart, the Supervisor 'session' will end. """


class FiniteStateMachine:
    """ This class implements a very simple behaviour of FiniteStateMachine based on a single event.
    A state is able to evaluate itself for transitions.

    Attributes are:
        - state: the current state of this state machine ;
        - instance: the current state instance ;
        - redeploy_mark: a status telling if a DEPLOYMENT state is pending.
    """

    def __init__(self, supvisors: Any) -> None:
        """ Reset the state machine and the internal context.

        :param supvisors: the Supvisors global structure
        """
        self.supvisors = supvisors
        self.state: SupvisorsStates = SupvisorsStates.OFF
        self.instance: AbstractState = OffState(supvisors)
        self.redeploy_mark: bool = False

    @property
    def logger(self) -> Logger:
        """ Return the Supvisors logger. """
        return self.supvisors.logger

    @property
    def context(self) -> Context:
        """ Return the Supvisors context structure. """
        return self.supvisors.context

    def next(self) -> None:
        """ Send the event to the state and transitions if possible.
        The state machine re-sends the event as long as it transitions.

        :return: None
        """
        # periodic check of start / stop jobs
        self.supvisors.starter.check()
        self.supvisors.stopper.check()
        # check state machine
        self.set_state(self.instance.next())

    def set_state(self, next_state: Optional[SupvisorsStates]) -> None:
        """ Update the current state of the state machine and transitions as long as possible.
        The transition can be forced, especially when getting the first Master state.

        :param next_state: the new state
        :return: None
        """
        # in the event of a Slave FSM, the master state may not be known yet, hence the test on next_state
        while next_state and next_state != self.state:
            # check that the transition is allowed
            # a Slave Supvisors will always follow the Master state
            if self.context.is_master and next_state not in self._Transitions[self.state]:
                self.logger.critical(f'FiniteStateMachine.set_state: unexpected transition from {self.state.name}'
                                     f' to {next_state.name}')
                break
            # exit the current state
            self.instance.exit()
            # assign the new state and publish SupvisorsStatus event internally and externally
            self.state = next_state
            self.logger.warn(f'FiniteStateMachine.set_state: Supvisors in {self.state.name}')
            # publish the new state
            self.supvisors.context.publish_state_modes({'fsm_state': self.state})
            # create the new state and enters it
            if self.context.is_master:
                self.instance = self._MasterStateInstances[self.state](self.supvisors)
            else:
                self.instance = self._SlaveStateInstances[self.state](self.supvisors)
            self.instance.enter()
            # evaluate current state
            next_state = self.instance.next()

    def on_timer_event(self, event: Payload) -> None:
        """ Periodic task used to check if remote Supvisors instances are still active.
        This is also the main event trigger of this state machine. """
        invalidated_identifiers, failed_processes = self.context.on_timer_event(event)
        self.handle_instance_failures(invalidated_identifiers, failed_processes)

    def handle_instance_failures(self, invalidated_identifiers: NameList,
                                 failed_processes: Set[ProcessStatus]) -> None:
        """ Upon failure of at least one Supvisors instance.

        :param invalidated_identifiers: the identifiers of the invalidated Supvisors instances.
        :param failed_processes: the processes in failure.
        :return: None.
        """
        self.logger.debug(f'FiniteStateMachine.handle_failures: invalidated_identifiers={invalidated_identifiers}'
                          f' process_failures={[process.namespec for process in failed_processes]}')
        if invalidated_identifiers:
            # inform Starter and Stopper
            # process_failures may be removed if already in their pipes
            self.supvisors.starter.on_instances_invalidation(invalidated_identifiers, failed_processes)
            self.supvisors.stopper.on_instances_invalidation(invalidated_identifiers, failed_processes)
            # deal with process_failures and isolation only if in DEPLOYMENT, OPERATION or CONCILIATION states
            if self.state in WORKING_STATES:
                # the Master fixes failures if any (can happen after an identifier invalidation, a process crash
                #   or a conciliation request)
                if self.context.is_master:
                    for process in failed_processes:
                        self.supvisors.failure_handler.add_default_job(process)
        # trigger remaining jobs in RunningFailureHandler
        if self.context.is_master:
            self.supvisors.failure_handler.trigger_jobs()
        # trigger FSM for global status re-evaluation
        # the Master may have been invalidated
        # process_failures could also positively impact the conflicts in the CONCILIATION state
        self.next()

    # Event handling methods
    def on_tick_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to refresh the data related to the Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param event: the tick event.
        :return: None.
        """
        self.context.on_tick_event(status, event)

    def on_discovery_event(self, event: Tuple) -> None:
        """ This event is used to add new Supvisors instances into the Supvisors system.
        No need to test if the discovery mode is enabled. This is managed in the internal communication layer.

        :param event: the discovery event.
        :return: None.
        """
        # When Supvisors is in discovery mode, new Supvisors instances may be added on-the-fly
        if self.context.on_discovery_event(event[0], event[1]):
            # a DEPLOYMENT will be requested if a new Supvisors instance has been inserted
            self.redeploy_mark = True

    def on_process_state_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to refresh the process data related to the event sent from the Supvisors instance.
        This event also triggers the application starter and/or stopper.

        :param status: the Supvisors instance that sent the event.
        :param event: the process event.
        :return: None.
        """
        process = self.context.on_process_state_event(status, event)
        # returned process may be None if the event is linked to an unknown or an isolated instance
        if process:
            # inform starter and stopper
            self.supvisors.starter.on_event(process, status.identifier)
            self.supvisors.stopper.on_event(process, status.identifier)
            # trigger an automatic (so master only) behaviour for a running failure
            # process crash triggered only if running failure strategy related to application
            # Supvisors does not replace Supervisor in the present matter (use autorestart if necessary)
            if self.context.is_master and process.crashed():
                # local variables to keep it readable
                strategy = process.rules.running_failure_strategy
                if strategy == RunningFailureStrategies.RESTART:
                    self.on_restart()
                elif strategy == RunningFailureStrategies.SHUTDOWN:
                    self.on_shutdown()
                else:
                    stop_strategy = strategy == RunningFailureStrategies.STOP_APPLICATION
                    restart_strategy = strategy == RunningFailureStrategies.RESTART_APPLICATION
                    # to avoid infinite application restart, exclude the case where process state is forced
                    # indeed the process state forced to FATAL can only happen during a starting sequence
                    # (no instance found) so retry is useless
                    if (stop_strategy or restart_strategy) and process.forced_state is None:
                        self.supvisors.failure_handler.add_default_job(process)

    def on_process_added_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to fill the internal structures when processes have been added on a Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param event: the process information.
        :return: None.
        """
        self.context.load_processes(status, [event])

    def on_process_removed_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to fill the internal structures when a process has been added on a Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param event: the process identification
        :return: None
        """
        self.context.on_process_removed_event(status, event)

    def on_process_disability_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to fill the internal structures when a process has been enabled or disabled
        on a Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param event: the process identification
        :return: None
        """
        self.context.on_process_disability_event(status, event)

    def on_state_event(self, status: SupvisorsInstanceStatus, event: Payload) -> None:
        """ This event is used to get the FSM state of the master Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param event: the state event.
        :return: None.
        """
        self.logger.debug(f'FiniteStateMachine.on_state_event: Supvisors={status.usage_identifier} sent {event}')
        # WARN: local instance is already up-to-date, could even be a step beyond
        #   so ignore the event if it is a local event
        ref_master = self.context.master_identifier
        ref_supvisors_state = self.context.supvisors_state
        # update the Supvisors instance states and modes
        self.context.on_instance_state_event(status, event)
        # check if there has been changes in Master and/or its state
        if ref_master != self.context.master_identifier:
            self.logger.info(f'FiniteStateMachine.on_state_event: new Master Supvisors={self.context.master_identifier}'
                             f' in {self.context.supvisors_state}')
            # if there has been a Master change and the local identifier is involved, the FSM type has to change,
            #   so it is required to go back to INITIALIZATION state
            if self.context.local_identifier in [ref_master, self.context.master_identifier]:
                self.set_state(SupvisorsStates.INITIALIZATION)
        elif ref_supvisors_state != self.context.supvisors_state:
            # the Master has transitioned to another state, so trigger the FSM
            self.next()

    def on_all_process_info(self, status: SupvisorsInstanceStatus, all_info: Optional[PayloadList]) -> None:
        """ This event is used to fill the internal structures with processes available on the Supvisors instance.

        :param status: the Supvisors instance that sent the event.
        :param all_info: all the processes' information.
        :return: None.
        """
        self.context.load_processes(status, all_info)

    def on_instance_failure(self, status: SupvisorsInstanceStatus) -> None:
        """ This event is received when a Supervisor proxy raised a failure.

        :param status: the Supvisors instance that sent the event.
        :return: None.
        """
        process_failures: Set[ProcessStatus] = self.context.on_instance_failure(status)
        self.handle_instance_failures([status.identifier], process_failures)

    def on_authorization(self, status: SupvisorsInstanceStatus, authorized: Optional[bool]) -> None:
        """ This event is used to finalize the port-knocking between Supvisors instances.
        When a new Supvisors instance comes in the group, back to DEPLOYMENT for a possible deployment.

        :param status: the Supvisors instance that sent the event.
        :param authorized: the authorization status as seen by the remote Supvisors instance.
        :return: None.
        """
        self.logger.debug(f'FiniteStateMachine.on_authorization: identifier={status.usage_identifier}'
                          f' authorized={authorized}')
        if self.context.on_authorization(status, authorized):
            # a new Supvisors instance comes in group
            # a DEPLOYMENT phase is considered as applications could not be fully started due to this missing instance
            # the idea of simply going back to INITIALIZATION is rejected as it would imply a re-synchronization
            if self.context.is_master:
                if self.state in WORKING_STATES:
                    # it may not be relevant to transition directly to DEPLOYMENT from here
                    # the DEPLOYMENT and CONCILIATION states are temporary and pending on actions to be completed
                    # so mark the context to remember that a re-DEPLOYMENT can be considered at OPERATION level
                    self.redeploy_mark = True
                    self.logger.info('FiniteStateMachine.on_authorization: defer re-DEPLOYMENT caused by new'
                                     f' Supvisors={status.usage_identifier}')

    def on_restart_sequence(self) -> None:
        """ This event is used to transition the state machine to the DEPLOYMENT state.

        :return: None
        """
        if self.context.is_master:
            self.redeploy_mark = Forced
        else:
            # re-route the command to Master
            self.supvisors.rpc_handler.send_restart_sequence(self.context.master_identifier)

    def on_restart(self) -> None:
        """ This event is used to transition the state machine to the RESTARTING state.

        :return: None
        """
        if self.context.is_master:
            self.set_state(SupvisorsStates.RESTARTING)
        else:
            if self.context.master_identifier:
                # re-route the command to Master
                self.supvisors.rpc_handler.send_restart_all(self.context.master_identifier)
            else:
                message = 'no Master instance to perform the Supvisors restart request'
                self.logger.error(f'FiniteStateMachine.on_restart: {message}')
                raise ValueError(message)

    def on_shutdown(self) -> None:
        """ This event is used to transition the state machine to the SHUTTING_DOWN state.

        :return: None
        """
        if self.context.is_master:
            self.set_state(SupvisorsStates.SHUTTING_DOWN)
        else:
            if self.context.master_identifier:
                # re-route the command to Master
                self.supvisors.rpc_handler.send_shutdown_all(self.context.master_identifier)
            else:
                message = 'no Master instance to perform the Supvisors restart request'
                self.logger.error(f'FiniteStateMachine.on_restart: {message}')
                raise ValueError(message)

    def on_end_sync(self, master_identifier: str) -> None:
        """ End the synchronization phase using the given Master or trigger an election.

        :param master_identifier: the identifier of the Master Supvisors instance selected by the user
        :return: None
        """
        if master_identifier:
            self.context.master_identifier = master_identifier
        else:
            self.context.elect_master()
        # re-evaluate the FSM
        self.next()

    # Map between state enumerations and classes
    _MasterStateInstances = {SupvisorsStates.OFF: OffState,
                             SupvisorsStates.INITIALIZATION: InitializationState,
                             SupvisorsStates.DISTRIBUTION: MasterDeploymentState,
                             SupvisorsStates.OPERATION: MasterOperationState,
                             SupvisorsStates.CONCILIATION: MasterConciliationState,
                             SupvisorsStates.RESTARTING: MasterRestartingState,
                             SupvisorsStates.SHUTTING_DOWN: MasterShuttingDownState,
                             SupvisorsStates.FINAL: FinalState}

    _SlaveStateInstances = {SupvisorsStates.OFF: OffState,
                            SupvisorsStates.INITIALIZATION: InitializationState,
                            SupvisorsStates.DISTRIBUTION: SlaveMainState,
                            SupvisorsStates.OPERATION: SlaveMainState,
                            SupvisorsStates.CONCILIATION: SlaveMainState,
                            SupvisorsStates.RESTARTING: SlaveRestartingState,
                            SupvisorsStates.SHUTTING_DOWN: SlaveShuttingDownState,
                            SupvisorsStates.FINAL: FinalState}

    # Transitions allowed between states
    _Transitions = {SupvisorsStates.OFF: [SupvisorsStates.INITIALIZATION],
                    SupvisorsStates.INITIALIZATION: [SupvisorsStates.DISTRIBUTION],
                    SupvisorsStates.DISTRIBUTION: [SupvisorsStates.INITIALIZATION,
                                                   SupvisorsStates.OPERATION,
                                                   SupvisorsStates.RESTARTING,
                                                   SupvisorsStates.SHUTTING_DOWN],
                    SupvisorsStates.OPERATION: [SupvisorsStates.CONCILIATION,
                                                SupvisorsStates.DISTRIBUTION,
                                                SupvisorsStates.INITIALIZATION,
                                                SupvisorsStates.RESTARTING,
                                                SupvisorsStates.SHUTTING_DOWN],
                    SupvisorsStates.CONCILIATION: [SupvisorsStates.OPERATION,
                                                   SupvisorsStates.INITIALIZATION,
                                                   SupvisorsStates.RESTARTING,
                                                   SupvisorsStates.SHUTTING_DOWN],
                    SupvisorsStates.RESTARTING: [SupvisorsStates.FINAL],
                    SupvisorsStates.SHUTTING_DOWN: [SupvisorsStates.FINAL],
                    SupvisorsStates.FINAL: []}



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================


from supvisors.web.viewimage import *

from .base import DummyHttpContext


def test_stats_image():
    """ Test the values set at construction. """
    image = StatsImage()
    assert image.contents is None
    # create a buffer
    contents = image.new_image()
    assert image.contents is not None
    assert image.contents is contents
    assert not contents.closed
    # create a buffer again
    image.new_image()
    assert contents is not None
    assert contents is not image.contents
    assert contents.closed
    assert not image.contents.closed


def test_host_instances():
    """ Test the values set at construction. """
    assert host_cpu_img is not None
    assert host_cpu_img.contents is None
    assert host_mem_img is not None
    assert host_mem_img.contents is None
    assert host_net_io_img is not None
    assert host_net_io_img.contents is None
    assert host_disk_io_img is not None
    assert host_disk_io_img.contents is None
    assert host_disk_usage_img is not None
    assert host_disk_usage_img.contents is None


def test_process_instances():
    """ Test the values set at construction. """
    assert process_cpu_img is not None
    assert process_cpu_img.contents is None
    assert process_mem_img is not None
    assert process_mem_img.contents is None


def test_image_view():
    """ Test the values set at construction. """
    # test creation
    image = StatsImage()
    view = ImageView(DummyHttpContext(), image)
    assert view.buffer is image
    # test render with an image having no contents
    response = view()
    headers = response['headers']
    assert headers['Content-Type'] == 'image/png'
    assert headers['Pragma'] == 'no-cache'
    assert headers['Cache-Control'] == 'no-cache'
    assert headers['Expires'] == 'Thu, 01 Jan 1970 00:00:00 GMT'
    assert b'' == response['body']
    # test render with an image having contents
    contents = image.new_image()
    contents.write(b'Dummy contents')
    response = view()
    headers = response['headers']
    assert headers['Content-Type'] == 'image/png'
    assert headers['Pragma'] == 'no-cache'
    assert headers['Cache-Control'] == 'no-cache'
    assert headers['Expires'] == 'Thu, 01 Jan 1970 00:00:00 GMT'
    assert response['body'] == b'Dummy contents'


def test_host_cpu_image_view():
    """ Test the HostCpuImageView class. """
    view = HostCpuImageView(DummyHttpContext())
    assert view.buffer is host_cpu_img


def test_host_memory_image_view():
    """ Test the HostMemoryImageView class. """
    view = HostMemoryImageView(DummyHttpContext())
    assert view.buffer is host_mem_img


def test_host_network_io_image_view():
    """ Test the HostNetworkIoImageView class. """
    view = HostNetworkIoImageView(DummyHttpContext())
    assert view.buffer is host_net_io_img


def test_host_disk_io_image_view():
    """ Test the HostDiskIoImageView class. """
    view = HostDiskIoImageView(DummyHttpContext())
    assert view.buffer is host_disk_io_img


def test_host_disk_usage_image_view():
    """ Test the HostDiskUsageImageView class. """
    view = HostDiskUsageImageView(DummyHttpContext())
    assert view.buffer is host_disk_usage_img


def test_process_cpu_image_view():
    """ Test the ProcessCpuImageView class. """
    view = ProcessCpuImageView(DummyHttpContext())
    assert view.buffer is process_cpu_img


def test_process_memory_image_view():
    """ Test the ProcessMemoryImageView class. """
    view = ProcessMemoryImageView(DummyHttpContext())
    assert view.buffer is process_mem_img


def test_supervisor_icon_image():
    """ Test the SupervisorIconImage class. """
    # test default
    view = SupervisorIconImage(DummyHttpContext())
    assert view.buffer is not None
    # test that a second instance would get the same object
    ref_buffer = view.buffer
    view = SupervisorIconImage(DummyHttpContext())
    assert view.buffer is not None
    assert view.buffer is ref_buffer


def test_software_icon_image():
    """ Test the SoftwareIconImage class. """
    # test default
    view = SoftwareIconImage(DummyHttpContext())
    assert view.buffer is None
    # load an image
    icon_path = os.path.join(os.path.dirname(supervisor.__file__), 'ui', 'images', 'icon.png')
    SoftwareIconImage.set_path(icon_path)
    view = SoftwareIconImage(DummyHttpContext())
    assert view.buffer is not None
    # test that a second instance would get the same object
    ref_buffer = view.buffer
    view = SoftwareIconImage(DummyHttpContext())
    assert view.buffer is not None
    assert view.buffer is ref_buffer



#!/usr/bin/python
# -*- coding: utf-8 -*-

# ======================================================================
# Copyright 2022 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from urllib.parse import quote

from supervisor.compat import as_string
from supervisor.web import MeldView
from supervisor.xmlrpc import Faults, RPCError


class MainTailView(MeldView):
    """ Adaptation of supervisor.web.TailView to display the logs of Supervisor. """

    def render(self):
        """ Rendering of the Main tail page. """
        supvisors = self.context.supervisord.supvisors
        form = self.context.form
        # get log limit to read
        limit = form.get('limit', '1024')
        limit = min(-1024, int(limit) * -1 if limit.isdigit() else -1024)
        # read Supervisor logs from RPC interface
        rpc_intf = supvisors.supervisor_data.supervisor_rpc_interface
        try:
            tail = rpc_intf.readLog(limit, 0)
        except RPCError as e:
            if e.code == Faults.NO_FILE:
                tail = 'No file for Supervisor'
            else:
                tail = f'ERROR: unexpected rpc fault [{e.code}] {e.text}'
        # generate page
        root = self.clone()
        root.findmeld('title').content('Supervisor tail')
        root.findmeld('tailbody').content(tail)
        url = f'maintail.html?limit={quote(str(abs(limit)))}'
        root.findmeld('refresh_anchor').attributes(href=url)
        return as_string(root.write_xhtmlstring())



# ======================================================================
# Copyright 2017 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import http.client
import socket
from unittest.mock import call, patch, Mock, DEFAULT

import pytest
from supervisor.xmlrpc import Faults

from supvisors.internal_com.supervisorproxy import *
from supvisors.rpcinterface import SupvisorsFaults
from supvisors.ttypes import SupvisorsInstanceStates


def test_internal_event_headers():
    """ Test the InternalEventHeaders enumeration. """
    expected = ['REQUEST', 'PUBLICATION', 'NOTIFICATION']
    assert [x.name for x in InternalEventHeaders] == expected


@pytest.fixture
def mocked_rpc():
    """ Fixture for the instance to test. """
    rpc_patch = patch('supvisors.internal_com.supervisorproxy.getRPCInterface')
    mocked_rpc = rpc_patch.start()
    yield mocked_rpc
    rpc_patch.stop()


@pytest.fixture
def proxy(supvisors, mocked_rpc):
    status = supvisors.context.instances['10.0.0.1:25000']
    return SupervisorProxy(status, supvisors)


@pytest.fixture
def proxy_thread(supvisors, mocked_rpc):
    status = supvisors.context.instances['10.0.0.1:25000']
    return SupervisorProxyThread(status, supvisors)


@pytest.fixture
def proxy_server(supvisors, mocked_rpc):
    server = SupervisorProxyServer(supvisors)
    yield server
    server.stop()


def test_proxy_creation(mocked_rpc, proxy, supvisors):
    """ Test the SupvisorsProxy creation. """
    assert proxy.supvisors is supvisors
    assert proxy.status is supvisors.context.instances['10.0.0.1:25000']
    assert proxy.proxy is not None
    assert 0.0 < proxy.last_used < time.monotonic()
    assert proxy.logger is supvisors.logger
    assert proxy.local_identifier == supvisors.mapper.local_identifier
    assert mocked_rpc.call_args_list == [call({'SUPERVISOR_SERVER_URL': 'http://10.0.0.1:25000',
                                               'SUPERVISOR_USERNAME': 'user',
                                               'SUPERVISOR_PASSWORD': 'p@$$w0rd'})]


def test_proxy_proxy(supvisors, proxy):
    """ Test the SupvisorsProxy proxy property. """
    # test with non-local proxy
    assert proxy.status.supvisors_id.identifier != proxy.local_identifier
    ref_proxy = proxy._proxy
    ref_usage = proxy.last_used
    assert proxy.proxy is ref_proxy
    assert proxy.last_used == ref_usage
    # test with local proxy and recent usage
    proxy.status = supvisors.context.local_status
    assert proxy.status.supvisors_id.identifier == proxy.local_identifier
    assert proxy.proxy is ref_proxy
    assert proxy.last_used == ref_usage
    # test with local proxy and old usage (cannot test everything due to patch)
    proxy.last_used = time.monotonic() - LOCAL_PROXY_DURATION - 1
    assert proxy.proxy
    assert proxy.last_used > ref_usage


def test_get_origin(supvisors, proxy):
    """ Test the SupervisorProxy._get_origin method. """
    local_instance = supvisors.mapper.local_instance
    assert proxy._get_origin(proxy.local_identifier) == local_instance.source
    assert proxy._get_origin('10.0.0.1:25000') == ('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000))


def test_proxy_xml_rpc(supvisors, proxy):
    """ Test the SupervisorProxy function to send any XML-RPC to a Supervisor instance. """
    ref_usage = proxy.last_used
    mocked_fct = Mock()
    # test no error
    proxy.xml_rpc('normal', mocked_fct, ())
    assert mocked_fct.call_args_list == [call()]
    assert proxy.last_used > ref_usage
    ref_usage = proxy.last_used
    mocked_fct.reset_mock()
    proxy.xml_rpc('normal', mocked_fct, ('hello',))
    assert mocked_fct.call_args_list == [call('hello')]
    assert proxy.last_used > ref_usage
    ref_usage = proxy.last_used
    mocked_fct.reset_mock()
    proxy.xml_rpc('normal', mocked_fct, ('hello', 28))
    assert mocked_fct.call_args_list == [call('hello', 28)]
    assert proxy.last_used > ref_usage
    ref_usage = proxy.last_used
    mocked_fct.reset_mock()
    # test minor exception (remote Supvisors instance is operational)
    mocked_fct.side_effect = RPCError(code=58)
    proxy.xml_rpc('normal', mocked_fct, ('hello', 28))
    assert mocked_fct.call_args_list == [call('hello', 28)]
    assert proxy.last_used > ref_usage
    ref_usage = proxy.last_used
    mocked_fct.reset_mock()
    # test major exception (remote Supvisors instance is NOT operational)
    for exc_class in [OSError, HTTPException, xmlrpclib.Fault(77, 'fault'), KeyError, ValueError, TypeError]:
        mocked_fct.side_effect = exc_class
        with pytest.raises(SupervisorProxyException):
            proxy.xml_rpc('normal', mocked_fct, ('hello',))
        assert mocked_fct.call_args_list == [call('hello')]
        assert proxy.last_used > ref_usage
        ref_usage = proxy.last_used
        mocked_fct.reset_mock()


def test_proxy_comm_event(mocker, supvisors, proxy):
    """ Test the SupervisorProxy function to send a remote communication event to a Supervisor instance. """
    mocked_call = mocker.patch.object(proxy, 'xml_rpc')
    # test with a mocked rpc interface
    proxy.send_remote_comm_event('supvisors', 'event data')
    assert mocked_call.call_args_list == [call('supervisor.sendRemoteCommEvent',
                                               proxy.proxy.supervisor.sendRemoteCommEvent,
                                               ('supvisors', '"event data"'))]


def test_publish(mocker, supvisors, proxy):
    """ Test the SupervisorProxy.publish method. """
    mocked_call = mocker.patch.object(proxy, 'xml_rpc')
    # update node states
    instance_status = supvisors.context.instances['10.0.0.1:25000']
    # test incorrect publication type
    proxy.publish('10.0.0.2:25000', (28, {'message': 'hello'}))
    assert not mocked_call.called
    # test TICK publication
    origin = supvisors.mapper.instances['10.0.0.2:25000'].source
    for state in SupvisorsInstanceStates:
        # actually, publish is not called with ISOLATED state, but it is dealt before
        proxy.publish('10.0.0.2:25000', (PublicationHeaders.TICK.value, {'message': 'TICK'}))
        instance_status._state = state
        expected_json = '[["10.0.0.2:25000", "10.0.0.2", ["10.0.0.2", 25000]], [0, {"message": "TICK"}]]'
        assert mocked_call.call_args_list == [call('supervisor.sendRemoteCommEvent',
                                                   proxy.proxy.supervisor.sendRemoteCommEvent,
                                                   (SUPVISORS_PUBLICATION, expected_json))]
        mocker.resetall()
    # test non-TICK publication with non-active state
    for state in [SupvisorsInstanceStates.UNKNOWN, SupvisorsInstanceStates.SILENT, SupvisorsInstanceStates.ISOLATED]:
        instance_status._state = state
        proxy.publish('10.0.0.2:25000', (PublicationHeaders.STATE.value, {'message': 'hello'}))
        assert not mocked_call.called
    # test non-TICK publication with active state
    for state in [SupvisorsInstanceStates.CHECKING, SupvisorsInstanceStates.CHECKED, SupvisorsInstanceStates.RUNNING]:
        instance_status._state = state
        proxy.publish('10.0.0.2:25000', (PublicationHeaders.STATE.value, {'message': 'STATE'}))
        expected_json = '[["10.0.0.2:25000", "10.0.0.2", ["10.0.0.2", 25000]], [7, {"message": "STATE"}]]'
        assert mocked_call.call_args_list == [call('supervisor.sendRemoteCommEvent',
                                                   proxy.proxy.supervisor.sendRemoteCommEvent,
                                                   (SUPVISORS_PUBLICATION, expected_json))]
        mocker.resetall()


def test_proxy_check_instance(mocker, supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy.check_instance method. """
    mocked_auth = mocker.patch.object(proxy, '_is_authorized', return_value=False)
    mocked_mode = mocker.patch.object(proxy, '_transfer_states_modes')
    mocked_info = mocker.patch.object(proxy, '_transfer_process_info')
    mocked_send = supvisors.rpc_handler.proxy_server.push_notification
    # test with no authorization
    proxy.check_instance()
    assert mocked_auth.call_args_list == [call()]
    assert not mocked_mode.called
    assert not mocked_info.called
    expected = NotificationHeaders.AUTHORIZATION.value, False
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]
    mocked_send.reset_mock()
    mocker.resetall()
    # test with authorization
    mocked_auth.return_value = True
    proxy.check_instance()
    assert mocked_auth.call_args_list == [call()]
    assert mocked_mode.call_args_list == [call()]
    assert mocked_info.call_args_list == [call()]
    expected = NotificationHeaders.AUTHORIZATION.value, True
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]


def test_proxy_is_authorized(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy._is_authorized method. """
    info_rpc = proxy.proxy.supvisors.get_instance_info
    # test with transport failure
    info_rpc.side_effect = OSError
    with pytest.raises(SupervisorProxyException):
        proxy._is_authorized()
    assert info_rpc.call_args_list == [call(proxy.local_identifier)]
    info_rpc.reset_mock()
    # test with XML-RPC application failure
    info_rpc.side_effect = RPCError(Faults.NOT_EXECUTABLE)
    assert proxy._is_authorized() is None
    assert info_rpc.call_args_list == [call(proxy.local_identifier)]
    info_rpc.reset_mock()
    # test with a mocked rpc interface
    info_rpc.side_effect = None
    # test with local Supvisors instance isolated by remote
    info_rpc.return_value = [{'statecode': SupvisorsInstanceStates.ISOLATED.value}]
    assert proxy._is_authorized() is False
    assert info_rpc.call_args_list == [call(proxy.local_identifier)]
    info_rpc.reset_mock()
    # test with local Supvisors instance not isolated by remote
    for state in [x for x in SupvisorsInstanceStates if x != SupvisorsInstanceStates.ISOLATED]:
        info_rpc.return_value = [{'statecode': state.value}]
        assert proxy._is_authorized() is True
        assert info_rpc.call_args_list == [call(proxy.local_identifier)]
        info_rpc.reset_mock()
    # test with local Supvisors instance not isolated by remote but returning an unknown state
    info_rpc.return_value = [{'statecode': 128}]
    assert proxy._is_authorized() is False
    assert info_rpc.call_args_list == [call(proxy.local_identifier)]


def test_proxy_transfer_process_info(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy._transfer_process_info method. """
    info_rpc = proxy.proxy.supvisors.get_all_local_process_info
    mocked_send = supvisors.rpc_handler.proxy_server.push_notification
    # test with transport failure
    info_rpc.side_effect = HTTPException
    with pytest.raises(SupervisorProxyException):
        proxy._transfer_process_info()
    assert info_rpc.call_args_list == [call()]
    assert not mocked_send.called
    info_rpc.reset_mock()
    # test with XML-RPC application failure
    info_rpc.side_effect = RPCError(Faults.ABNORMAL_TERMINATION)
    proxy._transfer_process_info()
    assert info_rpc.call_args_list == [call()]
    expected = NotificationHeaders.ALL_INFO.value, None
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]
    info_rpc.reset_mock()
    mocked_send.reset_mock()
    # test with a mocked rpc interface
    all_info = [{'name': 'dummy_1'}, {'name': 'dummy_2'}]
    info_rpc.side_effect = None
    info_rpc.return_value = all_info
    proxy._transfer_process_info()
    assert info_rpc.call_args_list == [call()]
    expected = NotificationHeaders.ALL_INFO.value, all_info
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]


def test_proxy_transfer_states_modes(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy._transfer_states_modes method. """
    info_rpc = proxy.proxy.supvisors.get_instance_info
    mocked_send = supvisors.rpc_handler.proxy_server.push_notification
    # test with transport failure
    info_rpc.side_effect = xmlrpclib.Fault
    with pytest.raises(SupervisorProxyException):
        proxy._transfer_states_modes()
    assert info_rpc.call_args_list == [call('10.0.0.1:25000')]
    assert not mocked_send.called
    mocked_rpc.reset_mock()
    # test with XML-RPC application failure
    info_rpc.side_effect = RPCError(Faults.NO_FILE, 'a file')
    proxy._transfer_states_modes()
    assert info_rpc.call_args_list == [call('10.0.0.1:25000')]
    expected = NotificationHeaders.STATE.value, None
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]
    info_rpc.reset_mock()
    mocked_send.reset_mock()
    # test with a mocked rpc interface
    instance_info = {'identifier': 'supvisors', 'node_name': '10.0.0.1', 'port': 25000, 'loading': 0,
                     'statecode': 3, 'statename': 'RUNNING',
                     'remote_time': 50, 'local_time': 60,
                     'sequence_counter': 28, 'process_failure': False,
                     'fsm_statecode': 6, 'fsm_statename': 'SHUTTING_DOWN',
                     'discovery_mode': True,
                     'master_identifier': '10.0.0.1',
                     'starting_jobs': False, 'stopping_jobs': True}
    info_rpc.side_effect = None
    info_rpc.return_value = [instance_info]
    proxy._transfer_states_modes()
    assert info_rpc.call_args_list == [call('10.0.0.1:25000')]
    expected = NotificationHeaders.STATE.value, {'fsm_statecode': 6,
                                                 'discovery_mode': True,
                                                 'master_identifier': '10.0.0.1',
                                                 'starting_jobs': False, 'stopping_jobs': True}
    assert mocked_send.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)), expected))]


def test_proxy_start_process(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to start a process handled by a remote Supervisor. """
    start_rpc = proxy.proxy.supvisors.start_args
    # test with JSON failure
    start_rpc.side_effect = KeyError
    with pytest.raises(SupervisorProxyException):
        proxy.start_process('dummy_process', 'extra args')
    assert start_rpc.call_args_list == [call('dummy_process', 'extra args', False)]
    start_rpc.reset_mock()
    # test with XML-RPC application failure
    start_rpc.side_effect = RPCError(Faults.ALREADY_STARTED)
    proxy.start_process('dummy_process', 'extra args')
    assert start_rpc.call_args_list == [call('dummy_process', 'extra args', False)]
    start_rpc.reset_mock()
    # test with a mocked rpc interface
    start_rpc.side_effect = None
    proxy.start_process('dummy_process', 'extra args')
    assert start_rpc.call_args_list == [call('dummy_process', 'extra args', False)]


def test_proxy_stop_process(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to stop a process handled by a remote Supervisor. """
    stop_rpc = proxy.proxy.supervisor.stopProcess
    # test with transport failure
    stop_rpc.side_effect = ConnectionResetError
    with pytest.raises(SupervisorProxyException):
        proxy.stop_process('dummy_process')
    assert stop_rpc.call_args_list == [call('dummy_process', False)]
    stop_rpc.reset_mock()
    # test with XML-RPC application failure
    stop_rpc.side_effect = RPCError(Faults.NOT_RUNNING)
    proxy.stop_process('dummy_process')
    assert stop_rpc.call_args_list == [call('dummy_process', False)]
    stop_rpc.reset_mock()
    # test with a mocked rpc interface
    stop_rpc.side_effect = None
    proxy.stop_process('dummy_process')
    assert stop_rpc.call_args_list == [call('dummy_process', False)]


def test_proxy_restart(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to restart a remote Supervisor. """
    restart_rpc = proxy.proxy.supervisor.restart
    # test with JSON failure
    restart_rpc.side_effect = TypeError
    with pytest.raises(SupervisorProxyException):
        proxy.restart()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with XML-RPC application failure
    restart_rpc.side_effect = RPCError(Faults.SHUTDOWN_STATE)
    proxy.restart()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with a mocked rpc interface
    restart_rpc.side_effect = None
    proxy.restart()
    assert restart_rpc.call_args_list == [call()]


def test_proxy_shutdown(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to shut down a remote Supervisor. """
    shutdown_rpc = proxy.proxy.supervisor.shutdown
    # test with JSON failure
    shutdown_rpc.side_effect = socket.gaierror
    with pytest.raises(SupervisorProxyException):
        proxy.shutdown()
    assert shutdown_rpc.call_args_list == [call()]
    shutdown_rpc.reset_mock()
    # test with XML-RPC application failure
    shutdown_rpc.side_effect = RPCError(SupvisorsFaults.BAD_SUPVISORS_STATE)
    proxy.shutdown()
    assert shutdown_rpc.call_args_list == [call()]
    shutdown_rpc.reset_mock()
    # test with a mocked rpc interface
    shutdown_rpc.side_effect = None
    proxy.shutdown()
    assert shutdown_rpc.call_args_list == [call()]


def test_proxy_restart_sequence(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to trigger the start_sequence of Supvisors. """
    restart_rpc = proxy.proxy.supvisors.restart_sequence
    # test with transport failure
    restart_rpc.side_effect = http.client.IncompleteRead
    with pytest.raises(SupervisorProxyException):
        proxy.restart_sequence()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with XML-RPC application failure
    restart_rpc.side_effect = RPCError(SupvisorsFaults.SUPVISORS_CONF_ERROR)
    proxy.restart_sequence()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with a mocked rpc interface
    restart_rpc.side_effect = None
    proxy.restart_sequence()
    assert restart_rpc.call_args_list == [call()]


def test_proxy_restart_all(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to restart Supvisors. """
    restart_rpc = proxy.proxy.supvisors.restart
    # test with transport failure
    restart_rpc.side_effect = http.client.CannotSendHeader
    with pytest.raises(SupervisorProxyException):
        proxy.restart_all()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with XML-RPC application failure
    restart_rpc.side_effect = RPCError(Faults.FAILED)
    proxy.restart_all()
    assert restart_rpc.call_args_list == [call()]
    restart_rpc.reset_mock()
    # test with a mocked rpc interface
    restart_rpc.side_effect = None
    proxy.restart_all()
    assert restart_rpc.call_args_list == [call()]


def test_proxy_shutdown_all(supvisors, mocked_rpc, proxy):
    """ Test the SupervisorProxy function to shut down Supvisors. """
    shutdown_rpc = proxy.proxy.supvisors.shutdown
    # test with transport failure
    shutdown_rpc.side_effect = ConnectionRefusedError
    with pytest.raises(SupervisorProxyException):
        proxy.shutdown_all()
    assert shutdown_rpc.call_args_list == [call()]
    shutdown_rpc.reset_mock()
    # test with XML-RPC application failure
    shutdown_rpc.side_effect = RPCError(Faults.UNKNOWN_METHOD)
    proxy.shutdown_all()
    assert shutdown_rpc.call_args_list == [call()]
    shutdown_rpc.reset_mock()
    # test with a mocked rpc interface
    shutdown_rpc.side_effect = None
    proxy.shutdown_all()
    assert shutdown_rpc.call_args_list == [call()]


def check_call(proxy, mocked_loop, method_name, request, args):
    """ Perform a main loop request and check what has been called. """
    # send request
    proxy.execute((request.value, args))
    # test mocked main loop
    for key, mocked in mocked_loop.items():
        if key == method_name:
            assert mocked.call_count == 1
            assert mocked.call_args == call(*args) if args else call()
            mocked.reset_mock()
        else:
            assert not mocked.called


def test_proxy_execute(mocker, proxy):
    """ Test the SupervisorProxy function to execute a deferred Supervisor request. """
    # patch main loop subscriber
    mocked_proxy = mocker.patch.multiple(proxy, check_instance=DEFAULT,
                                         start_process=DEFAULT, stop_process=DEFAULT,
                                         restart=DEFAULT, shutdown=DEFAULT, restart_sequence=DEFAULT,
                                         restart_all=DEFAULT, shutdown_all=DEFAULT)
    # test incorrect request type
    proxy.execute((28, 'error'))
    for mocked in mocked_proxy.values():
        assert not mocked.called
    # test check instance
    check_call(proxy, mocked_proxy, 'check_instance', RequestHeaders.CHECK_INSTANCE, None)
    # test start process
    check_call(proxy, mocked_proxy, 'start_process', RequestHeaders.START_PROCESS, ('dummy_process', 'extra args'))
    # test stop process
    check_call(proxy, mocked_proxy, 'stop_process', RequestHeaders.STOP_PROCESS, ('dummy_process',))
    # test restart
    check_call(proxy, mocked_proxy, 'restart', RequestHeaders.RESTART, None)
    # test shutdown
    check_call(proxy, mocked_proxy, 'shutdown', RequestHeaders.SHUTDOWN, None)
    # test restart_sequence
    check_call(proxy, mocked_proxy, 'restart_sequence', RequestHeaders.RESTART_SEQUENCE, None)
    # test restart_all
    check_call(proxy, mocked_proxy, 'restart_all', RequestHeaders.RESTART_ALL, None)
    # test shutdown
    check_call(proxy, mocked_proxy, 'shutdown_all', RequestHeaders.SHUTDOWN_ALL, None)


def test_proxy_thread_creation(mocked_rpc, proxy_thread, supvisors):
    """ Test the SupervisorProxyThread creation. """
    assert isinstance(proxy_thread, threading.Thread)
    assert isinstance(proxy_thread, SupervisorProxy)
    assert proxy_thread.supvisors is supvisors
    assert proxy_thread.logger is supvisors.logger
    assert proxy_thread.queue.empty()
    assert not proxy_thread.event.is_set()


def test_proxy_run_exception_local(mocker, supvisors, proxy_thread):
    """ Test the SupvisorsProxy thread run / stop with an exception on the local proxy. """
    mocked_publish = mocker.patch.object(proxy_thread, 'publish', side_effect=SupervisorProxyException)
    mocked_push = mocker.patch.object(supvisors.rpc_handler.proxy_server, 'push_notification')
    # start the thread
    proxy_thread.start()
    time.sleep(1)
    assert proxy_thread.is_alive()
    assert not proxy_thread.event.is_set()
    # force local identifier
    supvisors.mapper.local_identifier = proxy_thread.status.identifier
    # send a publication event that will raise an exception
    message = '10.0.0.2:25000', (PublicationHeaders.TICK.value, {'when': 1234})
    proxy_thread.push_message((InternalEventHeaders.PUBLICATION, message))
    time.sleep(1.0)
    assert not proxy_thread.is_alive()
    assert not mocked_push.called


def test_proxy_run_exception_remote_inactive(mocker, supvisors, proxy_thread):
    """ Test the SupvisorsProxy thread run / stop with an exception on a remote proxy
    in the event where the remote status is not considered active. """
    mocked_publish = mocker.patch.object(proxy_thread, 'publish', side_effect=SupervisorProxyException)
    mocked_push = mocker.patch.object(supvisors.rpc_handler.proxy_server, 'push_notification')
    # start the thread
    proxy_thread.start()
    time.sleep(1)
    assert proxy_thread.is_alive()
    assert not proxy_thread.event.is_set()
    # force local identifier to a non-running instance
    supvisors.mapper.local_identifier = '10.0.0.3:25000'
    assert not proxy_thread.status.has_active_state()
    # send a publication event that will raise an exception
    message = '10.0.0.2:25000', (PublicationHeaders.TICK.value, {'when': 1234})
    proxy_thread.push_message((InternalEventHeaders.PUBLICATION, message))
    time.sleep(1.0)
    assert mocked_publish.call_args_list == [call('10.0.0.2:25000', (0, {'when': 1234}))]
    assert not proxy_thread.is_alive()
    assert not mocked_push.called


def test_proxy_run_exception_remote_active(mocker, supvisors, proxy_thread):
    """ Test the SupvisorsProxy thread run / stop with an exception on a remote proxy
    in the event where the remote status is considered active. """
    mocked_publish = mocker.patch.object(proxy_thread, 'publish', side_effect=SupervisorProxyException)
    mocked_push = mocker.patch.object(supvisors.rpc_handler.proxy_server, 'push_notification')
    # start the thread
    proxy_thread.start()
    time.sleep(1)
    assert proxy_thread.is_alive()
    assert not proxy_thread.event.is_set()
    # force local identifier to a non-running instance
    supvisors.mapper.local_identifier = '10.0.0.3:25000'
    proxy_thread.status._state = SupvisorsInstanceStates.RUNNING
    assert proxy_thread.status.has_active_state()
    # send a publication event that will raise an exception
    message = '10.0.0.2:25000', (PublicationHeaders.TICK.value, {'when': 1234})
    proxy_thread.push_message((InternalEventHeaders.PUBLICATION, message))
    time.sleep(1.0)
    assert mocked_publish.call_args_list == [call('10.0.0.2:25000', (0, {'when': 1234}))]
    assert not proxy_thread.is_alive()
    assert mocked_push.call_args_list == [call((('10.0.0.1:25000', '10.0.0.1', ('10.0.0.1', 25000)),
                                                (NotificationHeaders.INSTANCE_FAILURE.value, None)))]


def test_proxy_run(mocker, proxy_thread):
    """ Test the SupvisorsProxy thread run / stop. """
    mocked_exec = mocker.patch.object(proxy_thread, 'execute')
    mocked_publish = mocker.patch.object(proxy_thread, 'publish')
    mocked_send = mocker.patch.object(proxy_thread, 'send_remote_comm_event')
    # start the thread
    proxy_thread.start()
    time.sleep(1)
    assert proxy_thread.is_alive()
    assert not proxy_thread.event.is_set()
    # send a publication event
    message = '10.0.0.2', (PublicationHeaders.TICK.value, {'when': 1234})
    proxy_thread.push_message((InternalEventHeaders.PUBLICATION, message))
    time.sleep(1.0)
    assert mocked_publish.call_args_list == [call(*message)]
    assert not mocked_exec.called
    assert not mocked_send.called
    mocked_publish.reset_mock()
    # send a discovery event
    message = (('identifier', ('10.0.0.2', 51243)), (InternalEventHeaders.NOTIFICATION.value, {'when': 4321}))
    proxy_thread.push_message((InternalEventHeaders.NOTIFICATION, message))
    time.sleep(1.0)
    assert mocked_send.call_args_list == [call(SUPVISORS_NOTIFICATION, message)]
    assert not mocked_publish.called
    assert not mocked_exec.called
    mocked_send.reset_mock()
    # send a request
    proxy_thread.push_message((InternalEventHeaders.REQUEST, ('10.0.0.2', (RequestHeaders.RESTART_ALL, ''))))
    time.sleep(1.0)
    assert mocked_exec.call_args_list == [call((RequestHeaders.RESTART_ALL, ''))]
    assert not mocked_publish.called
    assert not mocked_send.called
    mocked_exec.reset_mock()
    # stop the thread
    proxy_thread.stop()


def test_proxy_server_creation(supvisors, proxy_server):
    """ Test the SupervisorProxyServer creation. """
    assert proxy_server.supvisors is supvisors
    assert proxy_server.proxies == {}
    assert proxy_server.local_identifier == supvisors.mapper.local_identifier


def test_proxy_server_get_proxy(supvisors, mocked_rpc, proxy_server):
    """ Test the SupervisorProxyServer get_proxy method. """
    assert proxy_server.proxies == {}
    # get a proxy from a non-isolated instance (instance not stored)
    supvisors.context.instances['10.0.0.1:25000']._state = SupvisorsInstanceStates.RUNNING
    proxy_1 = proxy_server.get_proxy('10.0.0.1:25000')
    assert proxy_1 is not None
    assert proxy_1.status.identifier == '10.0.0.1:25000'
    assert proxy_server.proxies == {'10.0.0.1:25000': proxy_1}
    # get a proxy from a non-isolated instance (instance not stored / double test)
    supvisors.context.instances['10.0.0.2:25000']._state = SupvisorsInstanceStates.SILENT
    proxy_2 = proxy_server.get_proxy('10.0.0.2:25000')
    assert proxy_2 is not None
    assert proxy_2.status.identifier == '10.0.0.2:25000'
    assert proxy_server.proxies == {'10.0.0.1:25000': proxy_1, '10.0.0.2:25000': proxy_2}
    # get a proxy from an isolated instance (instance not stored)
    supvisors.context.instances['10.0.0.3:25000']._state = SupvisorsInstanceStates.ISOLATED
    proxy_3 = proxy_server.get_proxy('10.0.0.3:25000')
    assert proxy_3 is None
    assert proxy_server.proxies == {'10.0.0.1:25000': proxy_1, '10.0.0.2:25000': proxy_2}
    # get a proxy from a non-isolated instance (instance stored)
    proxy_1bis = proxy_server.get_proxy('10.0.0.1:25000')
    assert proxy_1bis is proxy_1
    # get a proxy from an isolated instance (instance stored)
    supvisors.context.instances['10.0.0.1:25000']._state = SupvisorsInstanceStates.ISOLATED
    proxy_1ter = proxy_server.get_proxy('10.0.0.1:25000')
    assert proxy_1ter is None
    assert proxy_server.proxies == {'10.0.0.2:25000': proxy_2}
    # get a proxy from a non-isolated instance (instance stored but thread closed)
    proxy_2.stop()
    proxy_2.join()
    proxy_2bis = proxy_server.get_proxy('10.0.0.2:25000')
    assert proxy_2bis is not proxy_2
    assert proxy_server.proxies == {'10.0.0.2:25000': proxy_2bis}
    # test stop
    proxy_server.stop()
    assert proxy_server.proxies == {'10.0.0.2:25000': proxy_2bis}
    assert not proxy_2bis.is_alive()


def test_server_proxy_push_message(mocker, supvisors, mocked_rpc, proxy_server):
    """ Test the SupervisorProxy function to send a remote communication event to a Supervisor instance. """
    mocked_push = mocker.patch('supvisors.internal_com.supervisorproxy.SupervisorProxyThread.push_message')
    test_identifier = f'{socket.getfqdn()}:15000'
    # get a proxy from a non-isolated instance (instance not stored)
    supvisors.context.local_status._state = SupvisorsInstanceStates.CHECKED
    supvisors.context.instances['10.0.0.1:25000']._state = SupvisorsInstanceStates.RUNNING
    supvisors.context.instances['10.0.0.2:25000']._state = SupvisorsInstanceStates.SILENT
    supvisors.context.instances['10.0.0.3:25000']._state = SupvisorsInstanceStates.ISOLATED
    supvisors.context.instances['10.0.0.4:25000']._state = SupvisorsInstanceStates.ISOLATED
    supvisors.context.instances['10.0.0.5:25000']._state = SupvisorsInstanceStates.CHECKING
    supvisors.context.instances[test_identifier]._state = SupvisorsInstanceStates.UNKNOWN
    assert proxy_server.proxies == {}
    local_identifier = supvisors.mapper.local_identifier
    # test request to a non-isolated Supervisor
    proxy_server.push_request('10.0.0.1:25000', {'message': 'test request'})
    assert mocked_push.call_args_list == [call((InternalEventHeaders.REQUEST,
                                                (local_identifier, {'message': 'test request'})))]
    assert sorted(proxy_server.proxies.keys()) == ['10.0.0.1:25000']
    mocked_push.reset_mock()
    # test request to an isolated Supervisor
    proxy_server.push_request('10.0.0.3:25000', {'message': 'test failed request'})
    assert not mocked_push.called
    assert sorted(proxy_server.proxies.keys()) == ['10.0.0.1:25000']
    # stop all and reset
    proxy_server.stop()
    proxy_server.proxies = {}
    # test publish
    proxy_server.push_publication({'message': 'test publish'})
    assert len(mocked_push.call_args_list) == 4
    for called in mocked_push.call_args_list:
        assert called == call((InternalEventHeaders.PUBLICATION, (local_identifier, {'message': 'test publish'})))
    assert sorted(proxy_server.proxies.keys()) == ['10.0.0.1:25000', '10.0.0.2:25000', '10.0.0.5:25000', test_identifier]
    mocked_push.reset_mock()
    # stop all and reset
    proxy_server.stop()
    proxy_server.proxies = {}
    # test post_discovery
    proxy_server.push_notification({'message': 'test discovery'})
    assert mocked_push.call_args_list == [call((InternalEventHeaders.NOTIFICATION, {'message': 'test discovery'}))]
    assert sorted(proxy_server.proxies.keys()) == [local_identifier]



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import time
from typing import AbstractSet, Any, Dict, Optional, Set, Tuple

from supervisor.loggers import Logger, LevelsByName
from supervisor.options import make_namespec
from supervisor.rpcinterface import SupervisorNamespaceRPCInterface
from supervisor.states import ProcessStates, getProcessStateDescription, RUNNING_STATES, STOPPED_STATES

from .ttypes import NameList, Payload, RunningFailureStrategies, StartingFailureStrategies
from .utils import WILDCARD


class ProcessRules:
    """ Defines the rules for starting a process, iaw rules file.

    Attributes are:
        - supvisors: the Supvisors global structure ;
        - identifiers: the identifiers of the Supvisors instances where the process can be started (default: all) ;
        - at_identifiers: when @ rule is used, the process can be started on only one Supvisors instance
          but multiple processes of the same homogeneous group can be started on the same Supvisors instance ;
        - hash_identifiers: when # rule is used, the processes of the same homogeneous group can only be started
          on one the Supvisors instances ;
        - start_sequence: the order in the starting sequence of the application ;
        - stop_sequence: the order in the stopping sequence of the application ;
        - required: a status telling if the process is required within the application ;
        - wait_exit: a status telling if Supvisors has to wait for the process to exit before triggering the next phase
          in the starting sequence of the application ;
        - expected_load: the expected loading of the process on the considered hardware (can be anything
          at the user discretion: CPU, RAM, etc.) ;
        - starting_failure_strategy: supersedes the application rule and defines the strategy to apply
          when the process crashes when the application is starting ;
        - running_failure_strategy: supersedes the application rule and defines the strategy to apply
          when the process crashes when the application is running.
    """

    # attributes
    supvisors: Any = None
    identifiers: NameList = None
    at_identifiers: NameList = None
    hash_identifiers: NameList = None
    start_sequence: int = 0
    stop_sequence: int = -1
    required: bool = False
    wait_exit: bool = False
    expected_load: int = 0
    starting_failure_strategy: StartingFailureStrategies = StartingFailureStrategies.ABORT
    running_failure_strategy: RunningFailureStrategies = RunningFailureStrategies.CONTINUE

    def __init__(self, supvisors: Any) -> None:
        """ Initialization of the attributes.

        :param supvisors: the global Supvisors structure.
        """
        # keep a reference to the Supvisors global structure
        self.supvisors = supvisors
        # attributes
        self.identifiers = [WILDCARD]
        self.at_identifiers = []
        self.hash_identifiers = []

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    def check_at_identifiers(self, namespec: str, is_pattern: bool) -> None:
        """ When '@' is used, it must be part within a program/pattern element.

        :param namespec: the namespec of the program considered.
        :param is_pattern: True is the rules were taken from a pattern element.
        :return: None.
        """
        if self.at_identifiers and not is_pattern:
            self.logger.error(f'ProcessRules.check_at_identifiers: {namespec} - at_identifiers reset'
                              ' because not part of a pattern')
            self.identifiers, self.at_identifiers = [WILDCARD], []

    def check_hash_identifiers(self, namespec: str, is_pattern: bool) -> None:
        """ When '#' is used, it must be part within a program/pattern element.

        :param namespec: the namespec of the program considered.
        :param is_pattern: True is the rules were taken from a pattern element.
        :return: None.
        """
        if self.hash_identifiers and not is_pattern:
            self.logger.error(f'ProcessRules.check_hash_identifiers: {namespec} - hash_identifiers reset'
                              ' because not part of a pattern')
            self.identifiers, self.hash_identifiers = [WILDCARD], []

    def check_sign_identifiers(self, namespec: str) -> None:
        """ Having both '#' and '@' in the same rule does not make sense.

        :param namespec: the namespec of the program considered.
        :return: None.
        """
        if self.at_identifiers and self.hash_identifiers:
            self.logger.error(f'ProcessRules.check_sign_identifiers: {namespec} - hash_identifiers reset'
                              ' because both at_identifiers and hash_identifiers are set')
            self.hash_identifiers = []

    def check_start_sequence(self, namespec: str) -> None:
        """ ProcessRules having required=True MUST have start_sequence > 0,
        so force required to False if start_sequence is not set.

        :param namespec: the namespec of the program considered.
        :return: None.
        """
        if self.required and self.start_sequence == 0:
            self.logger.error(f'ProcessRules.check_start_sequence: {namespec} - required forced to False'
                              ' because no start_sequence defined')
            self.required = False

    def check_stop_sequence(self, namespec: str) -> None:
        """ Check the stop_sequence value.
        If stop_sequence hasn't been set from the rules file, use the same value as start_sequence.

        :param namespec: the namespec of the program considered.
        :return: None.
        """
        if self.stop_sequence < 0:
            self.logger.trace(f'ProcessRules.check_stop_sequence: {namespec}'
                              f' - set stop_sequence to {self.start_sequence}')
            self.stop_sequence = self.start_sequence

    def check_autorestart(self, namespec: str) -> None:
        """ Disable autorestart when RunningFailureStrategies is related to applications.
        In these cases, Supvisors triggers behaviors that are different to Supervisor.

        :param namespec: the namespec of the program considered.
        :return: None.
        """
        if self.running_failure_strategy in [RunningFailureStrategies.STOP_APPLICATION,
                                             RunningFailureStrategies.RESTART_APPLICATION]:
            try:
                if self.supvisors.supervisor_data.autorestart(namespec):
                    self.supvisors.supervisor_data.disable_autorestart(namespec)
                    self.logger.warn(f'ProcessRules.check_autorestart: namespec={namespec} - autorestart disabled'
                                     f' due to running failure strategy {self.running_failure_strategy.name}')
            except KeyError:
                self.logger.debug(f'ProcessRules.check_autorestart: namespec={namespec} unknown to local Supervisor')

    def check_dependencies(self, namespec: str, is_pattern: bool) -> None:
        """ Update rules after they have been read from the rules file.

        :param namespec: the namespec of the program considered.
        :param is_pattern: True is the rules were taken from a pattern element.
        :return: None.
        """
        self.check_at_identifiers(namespec, is_pattern)
        self.check_hash_identifiers(namespec, is_pattern)
        self.check_sign_identifiers(namespec)
        self.check_start_sequence(namespec)
        self.check_stop_sequence(namespec)
        self.check_autorestart(namespec)

    def __str__(self) -> str:
        """ Get the process rules as string.

        :return: the printable process rules.
        """
        return (f'identifiers={self.identifiers}'
                f' at_identifiers={self.at_identifiers} hash_identifiers={self.hash_identifiers}'
                f' start_sequence={self.start_sequence} stop_sequence={self.stop_sequence} required={self.required}'
                f' wait_exit={self.wait_exit} expected_load={self.expected_load}'
                f' starting_failure_strategy={self.starting_failure_strategy.name}'
                f' running_failure_strategy={self.running_failure_strategy.name}')

    def serial(self) -> Payload:
        """ Get a serializable form of the process rules.
        hash_identifiers is not exported as used internally to resolve identifiers.

        :return: the process rules in a dictionary.
        """
        return {'identifiers': self.identifiers,
                'start_sequence': self.start_sequence, 'stop_sequence': self.stop_sequence,
                'required': self.required, 'wait_exit': self.wait_exit,
                'expected_loading': self.expected_load,
                'starting_failure_strategy': self.starting_failure_strategy.name,
                'running_failure_strategy': self.running_failure_strategy.name}


class ProcessStatus:
    """ Class defining the status of a process of Supvisors.

    Attributes are:
        - supvisors: the Supvisors global structure ;
        - logger: the Supvisors global logger ;
        - application_name: the application name, or group name from a Supervisor point of view ;
        - process_name: the process name ;
        - namespec: the process namespec ;
        - state: the synthetic state of the process, same enumeration as Supervisor ;
        - forced_state: the state forced by Supvisors upon unexpected event ;
        - forced_reason: the reason why the state would be forced ;
        - expected_exit: a status telling if the process has exited expectantly ;
        - last_event_time: the local monotonic time of the last information received ;
        - extra_args: the additional arguments passed to the command line ;
        - program_name: the program name, as found in the [program] section of the Supervisor configuration file ;
        - process_index: always 0 unless the process is part of a homogeneous group ;
        - running_identifiers: the list of all Supervisors where the process is running ;
        - info_map: a process info dictionary for each Supvisors identifier (running or not) ;
        - rules: the rules related to this process.
    """

    supvisors: Any = None
    # attributes
    application_name: str = ''
    process_name: str = ''
    namespec: str = ''
    _state: ProcessStates = ProcessStates.UNKNOWN
    forced_state: Optional[ProcessStates] = None
    forced_reason: str = ''
    expected_exit: bool = True
    last_event_time: float = 0.0
    # common information across all Supvisors instances
    _program_name: str = ''
    _process_index: int = 0
    _extra_args: str = ''
    # rules part
    rules: ProcessRules = None
    # only one running identifier is expected for managed processes
    running_identifiers: Set[str] = None
    info_map: Dict[str, Payload] = None

    def __init__(self, application_name: str, process_name: str, rules: ProcessRules, supvisors: Any) -> None:
        """ Initialization of the attributes.

        :param application_name: the name of the application the process belongs to
        :param process_name: the name of the process
        :param rules: the rules loaded from the rules file
        :param supvisors: the global Supvisors structure
        """
        # keep a reference of the Supvisors data
        self.supvisors = supvisors
        # attributes
        self.application_name = application_name
        self.process_name = process_name
        self.namespec = make_namespec(application_name, process_name)
        # rules part
        self.rules = rules
        # one single running identifier is expected for managed processes
        self.running_identifiers = set()  # identifiers
        self.info_map = {}  # identifier: process_info

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    @property
    def state(self) -> ProcessStates:
        """ Getter of state attribute.

        :return: the process state
        """
        return self._state

    @property
    def displayed_state(self) -> ProcessStates:
        """ Getter of state attribute for display to the user.
        Returns forced state in priority if set.

        :return: the process state
        """
        return self.state if self.forced_state is None else self.forced_state

    @state.setter
    def state(self, new_state: ProcessStates) -> None:
        """ Setter of state attribute.

        :param new_state: the new process state
        :return: None
        """
        if self._state != new_state:
            self._state = new_state
            self.logger.info(f'ProcessStatus.state: {self.namespec} is {self.state_string()}')

    def force_state(self, event: Payload) -> bool:
        """ Force the process state due to an unexpected event.
        This may be caused by a process command that has not been acknowledged in due time, so here is a final check
        in the event where messages have crossed.

        :param event: the forced event
        :return: True if the process state has been forced
        """
        force_state = True
        # check current state on targeted identifier
        # Note: identifier may correspond to the Master identifier when it did not find any candidate Supvisors instance
        #       to start the process and this process may not be configured in the Supervisor of the Master
        identifier = event['identifier']
        event_time = event['now_monotonic']
        if identifier in self.info_map:
            instance_info = self.info_map[identifier]
            force_state = instance_info['event_time'] <= event_time
        # apply forced state only if no event has been received since the forced state has been evaluated
        if force_state:
            self.last_event_time = time.monotonic()
            self.forced_state = event['state']
            self.forced_reason = event['spawnerr']
            self.logger.info(f'ProcessStatus.force_state: {self.namespec} is {self.displayed_state_string()}'
                             f' (real: {self.state_string()}) - ({self.forced_reason})')
        else:
            self.logger.debug(f'ProcessStatus.force_state: forced event dismissed for {self.namespec}')
        return force_state

    def reset_forced_state(self, state: ProcessStates = None):
        """ Reset forced_state upon reception of new information only if not STOPPED (default state in Supervisor).

        :param state: the new state (provided only when Supervisor information is added for the first time)
        :return: None
        """
        if self.forced_state is not None and state != ProcessStates.STOPPED:
            self.forced_state = None
            self.forced_reason = ''
            self.logger.debug(f'ProcessStatus.reset_forced_state: namespec={self.namespec}')

    @property
    def extra_args(self) -> str:
        """ Getter of the extra arguments attribute.

        :return: the extra arguments applicable to the command line of the process
        """
        return self._extra_args

    @extra_args.setter
    def extra_args(self, new_args: str) -> None:
        """ Setter of the extra arguments attribute.

        :param new_args: the extra arguments applicable to the command line of the process
        :return: None
        """
        if self._extra_args != new_args:
            self._extra_args = new_args
            try:
                self.supvisors.supervisor_data.update_extra_args(self.namespec, new_args)
            except KeyError:
                self.logger.debug(f'ProcessStatus.extra_args: namespec={self.namespec} unknown to local Supervisor')

    @property
    def program_name(self) -> str:
        """ Getter of the program_name attribute.

        :return: the program name
        """
        return self._program_name

    @program_name.setter
    def program_name(self, prg_name: str) -> None:
        """ Setter of the program_name attribute.

        :param prg_name: the new program name
        :return: None
        """
        if self._program_name and self._program_name != prg_name:
            # this is very unlikely, almost mischievous if it ever happens
            program_names = {identifier: info['program_name'] for identifier, info in self.info_map.items()}
            self.logger.error(f'ProcessStatus.program_name: inconsistent program_name for namespec={self.namespec}'
                              f' program_names={program_names}')
        # apply the latest received, whatever there is an issue or not
        self._program_name = prg_name

    @property
    def process_index(self) -> int:
        """ Getter of the _process_index attribute.

        :return: the process index (always 0 unless part of a homogeneous group)
        """
        return self._process_index

    @process_index.setter
    def process_index(self, idx: int) -> None:
        """ Setter of the _process_index attribute.

        :param idx: the new process index
        :return: None
        """
        if self._process_index > 0 and self._process_index != idx:
            # very strange but it may happen if numprocs_start options are configured differently
            proc_indexes = {identifier: info['process_index'] for identifier, info in self.info_map.items()}
            self.logger.error(f'ProcessStatus.process_index: inconsistent values for namespec={self.namespec}'
                              f' proc_indexes={proc_indexes}')
        # apply the latest received, whatever there is an issue or not
        self._process_index = idx

    def serial(self) -> Payload:
        """ Get a serializable form of the ProcessStatus.
        For the publication to an external subscriber, choose the displayed state.

        :return: the process status in a dictionary
        """
        return {'application_name': self.application_name,
                'process_name': self.process_name,
                'statecode': self.displayed_state, 'statename': self.displayed_state_string(),
                'expected_exit': self.expected_exit,
                'last_event_time': self.last_event_time,
                'identifiers': list(self.running_identifiers),
                'extra_args': self.extra_args}

    # access
    def get_pid(self, identifier: str) -> int:
        """ Return the PID of the process located on the Supvisors instance identified.

        :param identifier: the Supvisors instance identifier
        :return: the PID
        """
        if self.state == ProcessStates.RUNNING and identifier in self.running_identifiers:
            return self.info_map[identifier]['pid']
        return 0

    def disabled(self) -> bool:
        """ Check if the process is disabled on all Supvisors instances knowing the process.

        :return: the disabled status of the process
        """
        return self.info_map and all(info['disabled'] for info in self.info_map.values())

    def disabled_on(self, identifier: str) -> bool:
        """ Check if the process is disabled on the Supvisors instance identified.
        If there is no information about the Supvisors instance, it is not considered as disabled.

        :param identifier: the Supvisors instance identifier
        :return: the disabled status of the process on the considered Supvisors instance
        """
        return identifier in self.info_map and self.info_map[identifier]['disabled']

    def has_crashed(self) -> bool:
        """ Return True if any of the processes has ever crashed or has ever exited unexpectedly.

        :return: the crash status of the process
        """
        return any(info['has_crashed'] for info in self.info_map.values())

    def crashed(self, identifier: Optional[str] = None) -> bool:
        """ Return True if the process has crashed or has exited unexpectedly.

        :return: the crash status of the process
        """
        if identifier:
            info = self.info_map.get(identifier)
            if not info:
                return False
        else:
            info = {'state': self.state,
                    'expected': self.expected_exit}
        return ProcessStatus.is_crashed_event(info)

    @staticmethod
    def is_crashed_event(info: Payload) -> bool:
        """ Return True if the process has crashed or has exited unexpectedly.

        :return: the crash status of the process
        """
        return info['state'] == ProcessStates.FATAL or (info['state'] == ProcessStates.EXITED and not info['expected'])

    def stopped(self) -> bool:
        """ Return True if the process is stopped, as designed in Supervisor.

        :return: the stopped status of the process
        """
        return self.state in STOPPED_STATES

    def running(self) -> bool:
        """ Return True if the process is running, as designed in Supervisor.

        :return: the running status of the process
        """
        return self.state in RUNNING_STATES

    def running_on(self, identifier: str) -> bool:
        """ Check if the process is running on the Supvisors instance identified.

        :param identifier: the Supvisors instance identifier
        :return: the running status of the process on the considered Supvisors instance
        """
        return self.running() and identifier in self.running_identifiers

    def conflicting(self) -> bool:
        """ Check if the process is in a conflicting state (more than one instance running).

        :return: the conflict status of the process
        """
        return len(self.running_identifiers) > 1

    @staticmethod
    def update_description(info: Payload) -> str:
        """ Return the Supervisor way to describe the process status.

        :param info: a subset of the dict received from Supervisor.getProcessInfo.
        :return: the description of the process status
        """
        # IDE warning on first parameter but ignored as the Supervisor function should have been set as staticmethod
        return SupervisorNamespaceRPCInterface._interpretProcessInfo(None, info)

    # for display in Web UI
    def get_applicable_details(self) -> Tuple[Optional[str], str, bool, bool]:
        """ Get the process synthesis received from the process across all Supvisors instances.

        Priority is taken in the following order:
            1. the forced state ;
            2. the information coming from a Supervisor where the process is running ;
            3. the most recent information coming from a Supervisor where the process has been stopped.

        :return: the identifier of the Supervisor where the description comes, the process state description
        """
        self.logger.trace(f'ProcessStatus.get_applicable_details: START namespec={self.namespec}')
        identifier, desc, has_stdout, has_stderr = None, '', False, False
        # if the state is forced, return the reason why
        if self.forced_state is not None:
            self.logger.trace(f'ProcessStatus.get_applicable_details: namespec={self.namespec}'
                              f' - Supvisors=None [FORCED]description={self.forced_reason}')
            return identifier, self.forced_reason, has_stdout, has_stderr
        # search for process info where process is running
        info_map = dict(filter(lambda x: x[0] in self.running_identifiers, self.info_map.items()))
        if info_map:
            if len(info_map) == 1:
                # single instance
                identifier, info = next(iter(info_map.items()))
                nick_identifier = self.supvisors.mapper.get_nick_identifier(identifier)
                desc = info['description'] + ' on ' + nick_identifier
                has_stdout = info['has_stdout']
                has_stderr = info['has_stderr']
            else:
                # multiple instances. conflict
                identifiers = list(self.running_identifiers)
                nick_identifiers = [self.supvisors.mapper.get_nick_identifier(identifier)
                                    for identifier in identifiers]
                desc = f"conflict on {', '.join(nick_identifiers)}"
            self.logger.trace(f'ProcessStatus.get_applicable_details: namespec={self.namespec}'
                              f' - Supvisors={identifier} [running]description={desc}')
        else:
            # none running. sort info_map them by stop date to give the origin of the last stop
            identifier, info = max(self.info_map.items(), key=lambda x: x[1]['stop'])
            desc = info['description']
            # get additional info if the process has been started at least once
            if info['start'] > 0:
                nick_identifier = self.supvisors.mapper.get_nick_identifier(identifier)
                desc += ' on ' + nick_identifier
                has_stdout = info['has_stdout']
                has_stderr = info['has_stderr']
            self.logger.trace(f'ProcessStatus.get_description: namespec={self.namespec}'
                              f' - Supvisors={identifier} [stopped]description={desc}')
        return identifier, desc, has_stdout, has_stderr

    def has_stdout(self, identifier: str) -> bool:
        """ Consider that a process has stdout logs on a given Supvisors instance if configured so
        and if it has been started at least once.

        :param identifier: the identifier where the process is configured.
        :return: True if a stdout log file can be expected.
        """
        info = self.info_map[identifier]
        return info['has_stdout'] and info['start'] > 0

    def has_stderr(self, identifier: str) -> bool:
        """ Consider that a process has stderr logs on a given Supvisors instance if configured so
        and if it has been started at least once.

        :param identifier: the identifier where the process is configured.
        :return: True if a stderr log file can be expected.
        """
        info = self.info_map[identifier]
        return info['has_stderr'] and info['start'] > 0

    # rules consideration
    def possible_identifiers(self) -> NameList:
        """ Return the list of identifier where the program could be started.
        To achieve that, three conditions:
            - the Supervisor of the Supvisors instance must know the program ;
            - the Supvisors identifier must be declared in the rules file ;
            - the program shall not be disabled.

        :return: the list of identifiers where the program could be started
        """
        # get the applicable identifiers
        rules_identifiers = self.rules.identifiers
        self.logger.debug(f'ProcessStatus.possible_identifiers: rules_identifiers={rules_identifiers}')
        if WILDCARD in self.rules.identifiers:
            filtered_identifiers = list(self.supvisors.mapper.instances.keys())
        else:
            # filter identifiers based on currently known list (may change due to discovery mode)
            filtered_identifiers = self.supvisors.mapper.filter(rules_identifiers)
        self.logger.debug(f'ProcessStatus.possible_identifiers: filtered={filtered_identifiers}')
        self.logger.debug(f'ProcessStatus.possible_identifiers: info_map={list(self.info_map.keys())}')
        return [identifier
                for identifier in filtered_identifiers
                if identifier in self.info_map and not self.disabled_on(identifier)]

    # methods
    def state_string(self) -> str:
        """ Get the process state as a string.

        :return: the process state as a string
        """
        return getProcessStateDescription(self.state)

    def displayed_state_string(self) -> str:
        """ Get the displayed process state (considering forced state) as a string.

        :return: the displayed process state as a string
        """
        return getProcessStateDescription(self.displayed_state)

    def add_info(self, identifier: str, payload: Payload) -> None:
        """ Insert a new process information in internal list.

        :param identifier: the identifier of the Supvisors instance from which the information has been received
        :param payload: a subset of the dict received from Supervisor.getProcessInfo.
        :return: None
        """
        info = self.info_map[identifier] = payload
        # keep date of last information received
        # use local time here as there is no guarantee that Supvisors instances will be time-synchronized
        self.last_event_time = time.monotonic()
        info['local_time'] = self.last_event_time
        info['event_time'] = info['now_monotonic']
        self.update_uptime(info)
        # WARN: when a new Supvisors instance comes in group, extra_args is kept only if information ties in
        if self.extra_args != info['extra_args']:
            # difference of view, reset
            info['extra_args'] = ''
            self.extra_args = ''
        # store the program name and the process index at global level
        self.program_name = info['program_name']
        self.process_index = info['process_index']
        # keep history that the process has ever crashed in the Supvisors instance
        has_crashed = ProcessStatus.is_crashed_event(info)
        info['has_crashed'] = has_crashed | info.get('has_crashed', False)
        # log final payload
        self.logger.trace(f'ProcessStatus.add_info: namespec={self.namespec} - payload={info}'
                          f' added to Supvisors={identifier}')
        # process synthetic information
        self.reset_forced_state(info['state'])
        # update process status
        self.update_status(identifier, info['state'])

    def update_info(self, identifier: str, payload: Payload, external: bool = True) -> None:
        """ Update the internal process information with event payload.
        Event information is less complete than initial information.

        :param identifier: the identifier of the Supvisors instance from which the information has been received
        :param payload: the process information used to refresh internal data
        :param external: True if the event comes from another Supvisors instance
        :return: None
        """
        # refresh internal information
        info = self.info_map[identifier]
        self.logger.trace(f'ProcessStatus.update_info: namespec={self.namespec}'
                          f' - updating info[{identifier}]={info} with payload={payload}')
        info.update(payload)
        # keep date of last information received
        # use local time here as there is no guarantee that Supervisors will be time synchronized
        self.last_event_time = time.monotonic()
        info['local_time'] = self.last_event_time
        info['event_time'] = info['now_monotonic']
        # last received extra_args are always applicable
        self.extra_args = payload['extra_args']
        # complete missing information
        info['statename'] = getProcessStateDescription(info['state'])
        # re-evaluate description using Supervisor function
        info['description'] = self.update_description(info)
        # reset start time if process in a starting state
        new_state = info['state']
        if new_state in [ProcessStates.STARTING, ProcessStates.BACKOFF]:
            info['start'] = info['now']
            info['start_monotonic'] = info['now_monotonic']
        elif new_state == ProcessStates.FATAL:
            # NOTE: in supervisor, FATAL does not set stop, so do the same here
            pass
        elif new_state in STOPPED_STATES:
            info['stop'] = info['now']
            info['stop_monotonic'] = info['now_monotonic']
        self.update_uptime(info)
        # keep history that the process has ever crashed in the Supvisors instance
        if external:
            # typically, an instance invalidation will not impact the has_crashed status
            info['has_crashed'] |= ProcessStatus.is_crashed_event(info)
        # log final payload
        self.logger.debug(f'ProcessStatus.update_info: namespec={self.namespec} - new info[{identifier}]={info}')
        # process synthetic information
        self.reset_forced_state()
        # update / check running Supervisors
        self.update_status(identifier, new_state)

    def update_disability(self, identifier: str, disabled: bool) -> None:
        """ Update the disabled status of the process.

        :param identifier: the identifier of the Supvisors instance from which the disability has been received
        :param disabled: set to True if the process is disabled
        :return: None
        """
        if identifier in self.info_map:
            self.info_map[identifier]['disabled'] = disabled

    def update_times(self, identifier: str, remote_mtime: float, remote_time: float) -> None:
        """ Update the internal process information when a new tick is received from the remote Supvisors instance.

        :param identifier: the identifier of the Supvisors instance from which the tick has been received.
        :param remote_mtime: the TICK monotonic timestamp in the remote Supvisors instance time reference.
        :param remote_time: the TICK timestamp (seconds from Epoch) in the remote Supvisors instance time reference.
        :return: None.
        """
        info = self.info_map.get(identifier)
        if info:
            info['now_monotonic'] = remote_mtime
            info['now'] = remote_time
            self.update_uptime(info)
            # it may have an impact on the description depending on the process state
            info['description'] = self.update_description(info)

    @staticmethod
    def update_uptime(info: Payload) -> None:
        """ Update uptime entry of a process information.

        :param info: the process information related to a given Supervisor
        :return: None
        """
        info['uptime'] = ((info['now_monotonic'] - info['start_monotonic'])
                          if info['state'] in [ProcessStates.RUNNING, ProcessStates.STOPPING] else 0)

    def invalidate_identifier(self, identifier: str) -> bool:
        """ Update the status of a process that was running on a lost / non-responsive Supvisors instance.

        :param identifier: the identifier of the Supvisors instance from which no more information is received
        :return: True if process not running anywhere anymore
        """
        self.logger.debug(f'ProcessStatus.invalidate_identifier: namespec={self.namespec}'
                          f' - Supvisors={identifier} invalidated')
        failure = False
        if identifier in self.running_identifiers:
            # update process status with a FATAL payload
            info = self.info_map[identifier]
            payload = {'now': info['now'], 'now_monotonic': info['now_monotonic'],
                       'state': ProcessStates.FATAL,
                       'extra_args': info['extra_args'],
                       'expected': False,
                       'spawnerr': f'Supervisor {identifier} lost'}
            self.update_info(identifier, payload, False)
            failure = self.running_identifiers == set()
        return failure

    def remove_identifier(self, identifier: str) -> bool:
        """ Update the status of a process that has been removed from the identified Supvisors instance,
        following an XML-RPC update_numprocs.

        :param identifier: the identifier of the Supvisors instance from which the process has been removed
        :return: True if the process is not defined anywhere anymore
        """
        del self.info_map[identifier]
        return self.info_map == {}

    def update_status(self, identifier: str, new_state: ProcessStates) -> None:
        """ Updates the state and list of running Supvisors instances iaw the new event.

        :param identifier: the identifier of the Supvisors instance from which new information has been received
        :param new_state: the new state of the process on the considered Supervisor
        :return: None
        """
        # update running identifiers list
        if new_state in STOPPED_STATES:
            self.running_identifiers.discard(identifier)
        elif new_state in RUNNING_STATES:
            # replace if current state stopped-like, add otherwise
            if self.stopped():
                self.running_identifiers = {identifier}
            else:
                self.running_identifiers.add(identifier)
        # evaluate state iaw running identifiers
        if self.conflicting():
            self._evaluate_conflict()
        else:
            # no conflict so there's at most one element running
            if self.running_identifiers:
                self.state = self.info_map[list(self.running_identifiers)[0]]['state']
                self.expected_exit = True
            else:
                # priority set on STOPPING
                if any(info['state'] == ProcessStates.STOPPING for info in self.info_map.values()):
                    self.state = ProcessStates.STOPPING
                    self.expected_exit = True
                else:
                    # for STOPPED_STATES, consider the most recent event time in the local reference time
                    info = max(self.info_map.values(), key=lambda x: x['local_time'])
                    self.state = info['state']
                    self.expected_exit = info['expected']
        # log the new status
        if LevelsByName.TRAC >= self.logger.level:
            log_trace = f'ProcessStatus.update_status: namespec={self.namespec} is {self.state_string()}'
            if self.running_identifiers:
                log_trace += f' on {list(self.running_identifiers)}'
            self.logger.trace(log_trace)

    def _evaluate_conflict(self) -> None:
        """ Get a synthetic state if several processes are in a RUNNING-like state.

        :return: True if a conflict is detected, None otherwise
        """
        # several processes seem to be in a running state so that becomes tricky
        states = {self.info_map[identifier]['state'] for identifier in self.running_identifiers}
        self.logger.debug(f'ProcessStatus.evaluate_conflict: {self.process_name} multiple states'
                          f' {[getProcessStateDescription(x) for x in states]}'
                          f' for Supvisors={list(self.running_identifiers)}')
        # state synthesis done using the sorting of RUNNING_STATES
        self.state = self.running_state(states)

    @staticmethod
    def running_state(states: AbstractSet[int]) -> int:
        """ Return the first matching running state.
         The sequence defined in the Supervisor RUNNING_STATES is suitable here.

        :param states: a list of all process states of the present process over all Supervisors
        :return: a running state if found in list, STOPPING otherwise
        """
        # There may be STOPPING states in the list
        # In this state, the process is not removed yet from the running_identifiers
        return next((state for state in list(RUNNING_STATES) + [ProcessStates.STOPPING]
                     if state in states), ProcessStates.UNKNOWN)



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from supervisor.states import ProcessStates

from supvisors.application import ApplicationStatus
from supvisors.ttypes import PayloadList, Payload
from .viewcontext import *
from .viewhandler import ViewHandler
from .webutils import *


class ApplicationView(ViewHandler):
    """ Supvisors Application page. """

    def __init__(self, context):
        """ Call of the superclass constructors. """
        ViewHandler.__init__(self, context)
        self.page_name = APPLICATION_PAGE
        # init parameters
        self.application_name: str = ''
        self.application: Optional[ApplicationStatus] = None

    def handle_parameters(self) -> None:
        """ Retrieve the parameters selected on the web page.

        :return: None
        """
        ViewHandler.handle_parameters(self)
        # check if application name is available
        self.application_name = self.view_ctx.application_name
        try:
            # store application
            self.application = self.sup_ctx.applications[self.application_name]
        except KeyError:
            # may happen when the user clicks from a page of the previous launch while the current Supvisors is still
            # in INITIALIZATION stats or if wrong application_name set in URL
            self.logger.error(f'ApplicationView.handle_parameters: unknown application_name={self.application_name}')
            # redirect page to main page to avoid infinite error loop
            self.view_ctx.store_message = error_message(f'Unknown application: {self.application_name}')
            self.view_ctx.redirect = True

    def write_navigation(self, root):
        """ Rendering of the navigation menu with selection of the current application. """
        self.write_nav(root, appli=self.application_name, source=self.local_identifier)

    # RIGHT SIDE / HEADER part
    def write_status(self, header_elt):
        """ Rendering of the header part of the Supvisors Application page. """
        if self.application:
            # set application name
            header_elt.findmeld('application_mid').content(self.application_name)
            # set application state
            header_elt.findmeld('state_mid').content(self.application.state.name)
            # set LED iaw major/minor failures
            elt = header_elt.findmeld('state_led_mid')
            if self.application.major_failure:
                elt.attrib['class'] = 'status_red'
            elif self.application.minor_failure:
                elt.attrib['class'] = 'status_yellow'
            elif self.application.running():
                elt.attrib['class'] = 'status_green'
            else:
                elt.attrib['class'] = 'status_empty'

    def write_options(self, header_elt):
        """ Write application options. """
        self.write_starting_strategy(header_elt)
        if self.has_process_statistics:
            self.write_periods(header_elt)
        else:
            # hide the Statistics periods box
            header_elt.findmeld('period_div_mid').replace('')

    def write_starting_strategy(self, header_elt):
        """ Write applicable starting strategies. """
        # get the current strategy
        selected_strategy = self.view_ctx.strategy
        # set hyperlinks for strategy actions
        for strategy in StartingStrategies:
            button_elt = header_elt.findmeld('%s_div_mid' % strategy.name.lower())
            href_elt = button_elt.findmeld('%s_a_mid' % strategy.name.lower())
            if selected_strategy == strategy:
                update_attrib(button_elt, 'class', 'off active')
            else:
                update_attrib(button_elt, 'class', 'on')
                url = self.view_ctx.format_url('', self.page_name, **{STRATEGY: strategy.name})
                href_elt.attributes(href=url)

    def write_actions(self, header_elt):
        """ Write actions related to the application. """
        super().write_actions(header_elt)
        # configure start application button
        elt = header_elt.findmeld('startapp_a_mid')
        url = self.view_ctx.format_url('', self.page_name, **{ACTION: 'startapp'})
        elt.attributes(href=url)
        # configure stop application button
        elt = header_elt.findmeld('stopapp_a_mid')
        url = self.view_ctx.format_url('', self.page_name, **{ACTION: 'stopapp'})
        elt.attributes(href=url)
        # configure restart application button
        elt = header_elt.findmeld('restartapp_a_mid')
        url = self.view_ctx.format_url('', self.page_name, **{ACTION: 'restartapp'})
        elt.attributes(href=url)

    # RIGHT SIDE / BODY part
    def write_contents(self, contents_elt) -> None:
        """ Rendering of the contents part of the page.

        :param contents_elt: the root element of the page.
        :return: None.
        """
        if self.application:
            data = self.get_process_data()
            self.write_process_table(contents_elt, data)
            # check selected Process Statistics
            namespec = self.view_ctx.process_name
            identifier = self.view_ctx.identifier
            if namespec and identifier:
                status = self.view_ctx.get_process_status(namespec)
                if not status or status.stopped() or status.application_name != self.application_name:
                    self.logger.warn(f'ApplicationView.write_contents: unselect Process Statistics for {namespec}')
                    # form parameter is not consistent. remove it
                    self.view_ctx.process_name = ''
            # write selected Process Statistics
            namespec = self.view_ctx.process_name
            info = next(filter(lambda x: (x['namespec'] == namespec and x['identifier'] == identifier), data), {})
            self.write_process_statistics(contents_elt, info)

    def get_process_data(self) -> PayloadList:
        """ Collect sorted data on processes.

        :return: information about the application processes.
        """
        data = []
        for process in self.application.processes.values():
            namespec = process.namespec
            # add the process synthesis
            unexpected_exit = process.state == ProcessStates.EXITED and not process.expected_exit
            possible_identifiers = process.possible_identifiers()
            identifier, description, has_stdout, has_stderr = process.get_applicable_details()
            # get_applicable_details may return a list of identifiers
            nb_cores, proc_stats = 0, None
            if identifier:
                nb_cores, proc_stats = self.view_ctx.get_process_stats(namespec, identifier)
            data.append({'row_type': ProcessRowTypes.APPLICATION_PROCESS,
                         'application_name': process.application_name, 'process_name': process.process_name,
                         'namespec': namespec, 'identifier': identifier,
                         'disabled': process.disabled(), 'startable': len(possible_identifiers) > 0, 'stoppable': True,
                         'statename': process.displayed_state_string(), 'statecode': process.displayed_state,
                         'gravity': 'FATAL' if unexpected_exit else process.displayed_state_string(),
                         'has_crashed': process.has_crashed(),
                         'running_identifiers': list(process.running_identifiers),
                         'description': description,
                         'main': True, 'nb_items': len(process.info_map),
                         'expected_load': process.rules.expected_load,
                         'nb_cores': nb_cores, 'proc_stats': proc_stats,
                         'has_stdout': has_stdout, 'has_stderr': has_stderr})
            # add data depending on the process shex
            process_shex, _ = self.view_ctx.get_process_shex(process.process_name)
            if process_shex:
                # add the process details (name is not needed)
                # NOTE: start / restart actions are not allowed
                for identifier, info in process.info_map.items():
                    crashed = ProcessStatus.is_crashed_event(info)
                    nb_cores, proc_stats = self.view_ctx.get_process_stats(namespec, identifier)
                    data.append({'row_type': ProcessRowTypes.INSTANCE_PROCESS, 'namespec': namespec,
                                 'application_name': info['group'], 'process_name': '',
                                 'identifier': identifier,
                                 'disabled': info['disabled'], 'startable': False, 'stoppable': False,
                                 'statename': info['statename'], 'statecode': info['state'],
                                 'gravity': 'FATAL' if crashed else info['statename'],
                                 'has_crashed': info['has_crashed'],
                                 'running_identifiers': [identifier],  # used for button but not necessarily running
                                 'description': info['description'],
                                 'main': False, 'nb_items': 0,
                                 'expected_load': process.rules.expected_load,
                                 'nb_cores': nb_cores, 'proc_stats': proc_stats,
                                 'has_stdout': process.has_stdout(identifier),
                                 'has_stderr': process.has_stderr(identifier)})
        # re-arrange data using alphabetical order
        return sorted(data, key=lambda x: (x['namespec'], -x['row_type'].value, x['identifier']))

    def write_process_table(self, contents_elt, data: PayloadList):
        """ Rendering of the application processes managed through Supervisor.

        :param contents_elt: the root element of the page.
        :param data: the process data to be displayed.
        :return: None.
        """
        table_elt = contents_elt.findmeld('table_mid')
        if data:
            self.write_process_global_shex(table_elt)
            # remove stats columns if statistics are disabled
            self.write_common_process_table(table_elt)
            # loop on all processes
            shaded_proc_tr, shaded_detail_tr = False, False  # used to invert background style
            for tr_elt, info in table_elt.findmeld('tr_mid').repeat(data):
                # write common status (shared between this application view and node view)
                self.write_common_process_status(tr_elt, info)
                # print process name and running instances
                self.write_process(tr_elt, info)
                # specific updates
                if info['row_type'] == ProcessRowTypes.INSTANCE_PROCESS:
                    # remove shex td
                    tr_elt.findmeld('shex_td_mid').replace('')
                    # set line background and invert
                    apply_shade(tr_elt, shaded_detail_tr)
                    shaded_detail_tr = not shaded_detail_tr
                elif info['row_type'] == ProcessRowTypes.APPLICATION_PROCESS:
                    # print process shex
                    self.write_process_shex(tr_elt, info, shaded_proc_tr)
                    # set line background and invert
                    apply_shade(tr_elt, shaded_proc_tr)
                    shaded_proc_tr = not shaded_proc_tr
                    shaded_detail_tr = shaded_proc_tr
        else:
            table_elt.replace('No programs to display')

    def write_process_global_shex(self, table_elt) -> None:
        """ Write global shrink / expand buttons.

        :param table_elt: the process table element.
        :return: None.
        """
        shex = self.view_ctx.process_shex
        expand_shex = self.view_ctx.get_default_process_shex(self.application_name, True)
        shrink_shex = self.view_ctx.get_default_process_shex(self.application_name, False)
        self.write_global_shex(table_elt, PROC_SHRINK_EXPAND, shex, expand_shex, shrink_shex)

    def write_process_shex(self, tr_elt, info: Payload, shaded_tr: bool):
        """ Write the application process section into a table. """
        elt = tr_elt.findmeld('shex_td_mid')
        process_name = info['process_name']
        process_shex, inverted_shex = self.view_ctx.get_process_shex(process_name)
        self.logger.trace(f'ApplicationView.write_process_shex: process_name={process_name}'
                          f' process_shex={process_shex} inverted_shex={inverted_shex}')
        if process_shex:
            elt.attrib['rowspan'] = str(info['nb_items'] + 1)
            apply_shade(elt, shaded_tr)
        elt = elt.findmeld('shex_a_mid')
        elt.content(f'{SHEX_SHRINK if process_shex else SHEX_EXPAND}')
        url = self.view_ctx.format_url('', self.page_name, **{PROC_SHRINK_EXPAND: inverted_shex})
        elt.attributes(href=url)

    def write_process(self, tr_elt, info):
        """ Rendering of the cell corresponding to the process running instances. """
        running_a_elt = tr_elt.findmeld('running_a_mid')
        running_identifiers = info['running_identifiers']
        if len(running_identifiers) == 0:
            # no running instance: empty cell
            running_a_elt.replace('')
        else:
            running_span_elt = running_a_elt.findmeld('running_span_mid')
            if len(running_identifiers) == 1:
                # one running instance: use button to navigate to the process page
                identifier = running_identifiers[0]
                nick_identifier = self.supvisors.mapper.get_nick_identifier(identifier)
                running_span_elt.content(nick_identifier)
                url = self.view_ctx.format_url(identifier, PROC_INSTANCE_PAGE)
                running_a_elt.attributes(href=url)
            else:
                # multiple running instances: conflict
                running_span_elt.content('Conciliate')
                update_attrib(running_span_elt, 'class', 'blink')
                url = self.view_ctx.format_url('', CONCILIATION_PAGE)
                running_a_elt.attributes(href=url)

    # ACTIONS
    def make_callback(self, namespec: str, action: str):
        """ Triggers processing iaw action requested. """
        if self.application:
            # get current strategy
            strategy = self.view_ctx.strategy
            if action == 'startapp':
                return self.start_application_action(strategy)
            if action == 'stopapp':
                return self.stop_application_action()
            if action == 'restartapp':
                return self.restart_application_action(strategy)
            if namespec:
                if self.view_ctx.get_process_status(namespec) is None:
                    return delayed_error(f'No such process named {namespec}')
                if action == 'start':
                    return self.start_process_action(strategy, namespec)
                if action == 'stop':
                    return self.stop_process_action(namespec)
                if action == 'restart':
                    return self.restart_process_action(strategy, namespec)
                if action == 'clearlog':
                    return self.clearlog_process_action(namespec)
        return delayed_error('No application selected')

    # Application actions
    def start_application_action(self, strategy: StartingStrategies) -> Callable:
        """ Start the application iaw the strategy.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param strategy: the strategy to apply for starting the application.
        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('start_application', (strategy.value, self.application_name, wait),
                                         f'Application {self.application_name} started')

    def restart_application_action(self, strategy: StartingStrategies) -> Callable:
        """ Restart the application iaw the strategy.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param strategy: the strategy to apply for restarting the application.
        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('restart_application', (strategy.value, self.application_name, wait),
                                         f'Application {self.application_name} restarted')

    def stop_application_action(self) -> Callable:
        """ Stop the application.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('stop_application', (self.application_name, wait),
                                         f'Application {self.application_name} stopped')

    # Process actions
    def start_process_action(self, strategy: StartingStrategies, namespec: str) -> Callable:
        """ Start the process named namespec iaw the strategy.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param strategy: the strategy to apply for starting the process.
        :param namespec: the process namespec.
        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('start_process', (strategy.value, namespec, '', wait),
                                         f'Process {namespec} started')

    def restart_process_action(self, strategy: StartingStrategies, namespec: str) -> Callable:
        """ Restart the process named namespec iaw the strategy.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param strategy: the strategy to apply for restarting the process.
        :param namespec: the process namespec.
        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('restart_process', (strategy.value, namespec, '', wait),
                                         f'Process {namespec} restarted')

    def stop_process_action(self, namespec: str) -> Callable:
        """ Stop the process named namespec.
        The RPC wait parameter is linked to the auto-refresh parameter of the page.

        :param namespec: the process namespec.
        :return: a callable for deferred result.
        """
        wait = not self.view_ctx.auto_refresh
        return self.supvisors_rpc_action('stop_process', (namespec, wait), f'Process {namespec} stopped')

    def clearlog_process_action(self, namespec: str) -> Callable:
        """ Can't call supervisor StatusView source code from application view.
        Just do the same job.

        :param namespec: the process namespec.
        :return: a callable for deferred result.
        """
        return self.supervisor_rpc_action('clearProcessLogs', (namespec,), f'Log for {namespec} cleared')



# ======================================================================
# Copyright 2022 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import asyncio
import json
from typing import List, Optional

# timeout for async operations, in seconds
ASYNC_TIMEOUT = 1.0


# Common functions
def payload_to_bytes(message) -> bytes:
    """ Use a JSON serialization and encode the message with UTF-8.

    :param message: the message to encode.
    :return: the message serialized and encoded, as bytes.
    """
    return json.dumps(message).encode('utf-8')


def bytes_to_payload(message: bytes) -> List:
    """ Decode and de-serialize the bytearray into a message header and body.

    :param message: the message as bytes
    :return: the message header and body
    """
    return json.loads(message.decode('utf-8'))


async def read_stream(reader: asyncio.StreamReader) -> Optional[bytes]:
    """ Read a message from an asyncio StreamReader.

    :param reader: the asyncio StreamReader.
    :return: None if reader is closed, empty bytes if nothing to read, or the message as bytes.
    """
    try:
        # read the message size
        msg_size_as_bytes = await asyncio.wait_for(reader.readexactly(4), 1.0)
    except asyncio.TimeoutError:
        # nothing to read
        return b''
    except (asyncio.IncompleteReadError, ConnectionResetError):
        # socket closed during read operation read interruption
        return None
    # decode the message size
    msg_size = int.from_bytes(msg_size_as_bytes, byteorder='big')
    if msg_size == 0:
        return None
    try:
        # read the message itself
        msg_as_bytes = await asyncio.wait_for(reader.readexactly(msg_size), 1.0)
    except (asyncio.TimeoutError, asyncio.IncompleteReadError, ConnectionResetError):
        # unexpected message without body or socket closed during read operation
        return None
    # return the non-decoded message
    return msg_as_bytes


async def write_stream(writer: asyncio.StreamWriter, msg_as_bytes: bytes) -> bool:
    """ Write a message to an asyncio StreamWriter.

    :param writer: the asyncio StreamReader.
    :param msg_as_bytes: the message as bytes.
    :return: True if no exception.
    """
    buffer = len(msg_as_bytes).to_bytes(4, 'big') + msg_as_bytes
    try:
        writer.write(buffer)
        await writer.drain()
    except ConnectionResetError:
        return False
    return True



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import sys

from supervisor import supervisord
from supervisor.loggers import Logger, getLogger, handle_file, handle_stdout
from supervisor.supervisord import Supervisor

from supvisors.internal_com.mapper import SupvisorsMapper
from .commander import Starter, Stopper, StarterModel
from .context import Context
from .listener import SupervisorListener
from .options import SupvisorsOptions, SupvisorsServerOptions, Automatic, get_logger_configuration
from .sparser import Parser
from .statemachine import FiniteStateMachine
from .statscompiler import HostStatisticsCompiler, ProcStatisticsCompiler
from .strategy import RunningFailureHandler
from .supervisordata import SupervisorData
from .supervisorupdater import SupervisorUpdater
from .ttypes import Payload

# use ';' in logger output as separator because easier to cut
LOGGER_FORMAT = '%(asctime)s;%(levelname)s;%(message)s\n'


def create_logger(supervisor: Supervisor, logger_config: Payload) -> Logger:
    """ Create the logger that will be used in Supvisors.

    If logfile is not set or set to AUTO, Supvisors will use the Supervisor logger.
    Else Supvisors will log in the file defined in option.
    """
    # prefix all log traces with the software name if set
    prefix = logger_config['prefix']
    logger_format = f'{prefix};{LOGGER_FORMAT}' if prefix else LOGGER_FORMAT
    # configure the Supervisor logger
    logfile = logger_config['logfile']
    if logfile is Automatic:
        # use Supervisord logger but patch format anyway
        logger = supervisor.options.logger
        for handler in logger.handlers:
            handler.setFormat(logger_format)
        return logger
    # else create own Logger using Supervisor functions
    nodaemon = supervisor.options.nodaemon
    silent = supervisor.options.silent
    logger = getLogger(logger_config['loglevel'])
    # tag the logger so that it is properly closed when exiting
    # when not tagged, Supervisor closes it
    logger.SUPVISORS = True
    if nodaemon and not silent:
        handle_stdout(logger, logger_format)
    handle_file(logger, logfile, logger_format,
                rotating=not not logger_config['logfile_maxbytes'],
                maxbytes=logger_config['logfile_maxbytes'],
                backups=logger_config['logfile_backups'])
    return logger


class Supvisors:
    """ The Supvisors class used as a global structure passed to most Supvisors objects. """

    def __init__(self, supervisor: Supervisor, **config) -> None:
        """ Instantiation of all the Supvisors objects.

        :param supervisor: the Supervisor global structure
        """
        # WARN: The Supvisors communication objects cannot be created at this level.
        #       Before running, Supervisor forks when daemonized and the sockets would be lost.
        self.rpc_handler = None
        self.discovery_handler = None
        self.external_publisher = None
        # create logger using option from config
        logger_config = get_logger_configuration(**config)
        self.logger = create_logger(supervisor, logger_config)
        # get options from config
        self.options = SupvisorsOptions(supervisor, self.logger, **config)
        # re-realize configuration to get process configuration not stored within Supervisor options
        self.server_options = SupvisorsServerOptions(self)
        self.server_options.realize(sys.argv[1:], doc=supervisord.__doc__)
        # configure supervisor data wrapper
        self.supervisor_data = SupervisorData(self, supervisor)
        self.supervisor_updater = SupervisorUpdater(self)
        # get declared Supvisors instances and check local identifier
        self.mapper = SupvisorsMapper(self)
        try:
            self.mapper.configure(self.options.supvisors_list,
                                  self.options.stereotypes,
                                  self.options.core_identifiers)
        except ValueError:
            self.logger.critical('Supvisors: Wrong Supvisors configuration (supvisors_list)')
            raise
        # create statistics handler
        self.stats_collector = None
        try:
            from .statscollector import StatisticsCollectorProcess
            self.stats_collector = StatisticsCollectorProcess(self)
        except (ImportError, ModuleNotFoundError):
            self.logger.info('Supvisors: psutil not installed')
            self.logger.warn('Supvisors: this Supvisors instance cannot not collect statistics')
        self.host_compiler = HostStatisticsCompiler(self)
        self.process_compiler = ProcStatisticsCompiler(self.options, self.logger)
        # create context data
        self.context = Context(self)
        # create application starter and stopper
        self.starter = Starter(self)
        self.stopper = Stopper(self)
        # create application starter model
        self.starter_model = StarterModel(self)
        # create the failure handler of crashing processes
        # WARN: must be created before the state machine
        self.failure_handler = RunningFailureHandler(self)
        # check rules files
        try:
            self.parser = Parser(self)
        except Exception as exc:
            self.logger.warn(f'Supvisors: cannot parse rules files: {self.options.rules_files} - {exc}')
            self.parser = None
        # create Supervisor event subscriber
        self.listener = SupervisorListener(self)
        # create state machine
        self.fsm = FiniteStateMachine(self)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# ======================================================================
# Copyright 2017 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================


from io import StringIO

# Contents of a minimal Supervisor configuration file without Supvisors
NoSupvisors = StringIO('''
[inet_http_server]
port=:60000

[supervisord]
''')

# Contents of a minimal Supervisor configuration file including program definitions
ProgramConfiguration = StringIO('''
[inet_http_server]
port=:60000

[supervisord]

[group:dummy_group]
programs=dummy,dummies

[program:dummy]
command=ls

[program:dummies]
command=ls
process_name=dummy_%(process_num)d
numprocs=3

[fcgi-program:dumber]
command=ls
process_name=dumber_%(process_num)d
numprocs=2
numprocs_start=10
socket=tcp://localhost:9002

[eventlistener:dummy_ears]
command=ls
process_name=dummy_ears_%(process_num)d
events=EVENT
buffer_size=10
numprocs=2
numprocs_start=20
''')

# Contents of a rules file including schema errors
InvalidXmlTest = b'''\
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<root>
    <alias name="not used">nodes_prg_B3, nodes_appli_D</alias>
    <alias name="not used too">#, 10.0.0.1, 192.168.12.20</alias>
    <alias name="nodes_prg_B1">#</alias>
    <alias name="nodes_prg_B3">*, 10.0.0.4, 192.168.12.20</alias>
    <alias name="nodes_appli_D"> </alias>

    <model name="dummy_model_01">
        <stop_sequence>0</stop_sequence>
        <required>false</required>
        <wait_exit>false</wait_exit>
        <expected_loading>25</expected_loading>
        <running_failure_strategy>RESTART</running_failure_strategy>
    </model>

    <model name="dummy_model_02">
        <identifiers>#</identifiers>
        <start_sequence>1</start_sequence>
        <required>true</required>
        <wait_exit>true</wait_exit>
    </model>

    <model name="dummy_model_03">
        <identifiers>10.0.0.4, 10.0.0.2</identifiers>
        <stop_sequence>100</stop_sequence>
        <expected_loading>10</expected_loading>
    </model>

    <model name="dummy_model_04">
        <reference>dummy_model_01</reference>
    </model>

    <application name="dummy_application_A">
        <operational_status>'dummy</operational_status>
    </application>

    <application name="dummy_application_B">
        <starting_failure_strategy>STOP</starting_failure_strategy>
        <running_failure_strategy>RESTART_PROCESS</running_failure_strategy>
        <distribution>non</distribution>
        <start_sequence>1</start_sequence>
        <stop_sequence>4</stop_sequence>
        <operational_status>'dummy_program_B0' ; 'dummy_program_B2'</operational_status>

        <programs>
            <program name="dummy_program_B0"/>

            <program name="dummy_program_B1">
                <identifiers>nodes_prg_B1</identifiers>
                <start_sequence>3</start_sequence>
                <stop_sequence>50</stop_sequence>
                <required>true</required>
                <wait_exit>false</wait_exit>
                <expected_loading>5</expected_loading>
                <running_failure_strategy>CONTINUE</running_failure_strategy>
             </program>
    
            <program name="dummy_program_B2">
                <identifiers>10.0.0.3</identifiers>
                <required>true</required>
                <expected_loading>-1</expected_loading>
                <running_failure_strategy>RESTART_PROCESS</running_failure_strategy>
            </program>

            <program name="dummy_program_B3">
                <identifiers>nodes_prg_B3</identifiers>
                <required>false</required>
                <expected_loading>100</expected_loading>
                <running_failure_strategy>STOP_APPLICATION</running_failure_strategy>
            </program>
    
            <program name="dummy_program_B4">
                <identifiers>10.0.0.1, 10.0.0.2</identifiers>
                <start_sequence>-1</start_sequence>
                <stop_sequence>-2</stop_sequence>
                <required>28</required>
                <wait_exit>77</wait_exit>
                <expected_loading>-1</expected_loading>
                <running_failure_strategy>RESTART_APPLICATION</running_failure_strategy>
            </program>

            <program name="dummy_program_B5">
                <identifiers>10.0.0.3, 10.0.0.1, 10.0.0.5</identifiers>
                <start_sequence>start</start_sequence>
                <stop_sequence>stop</stop_sequence>
                <required>req</required>
                <wait_exit>wait</wait_exit>
                <expected_loading>fifty</expected_loading>
                <running_failure_strategy>BACK</running_failure_strategy>
            </program>
        </programs>

    </application>

    <application pattern="_C">
        <distribution>SINGLE_INSTANCE</distribution>
        <identifiers>192.256.16.10,*</identifiers>
        <start_sequence>20</start_sequence>
        <stop_sequence>0</stop_sequence>
        <starting_failure_strategy>ABORT</starting_failure_strategy>
        <running_failure_strategy>STOP_APPLICATION</running_failure_strategy>

        <programs>
            <program name="dummy_program_C0">
                <reference></reference>
            </program>
    
            <program name="dummy_program_C1">
                <reference>unknown</reference>
            </program>
    
            <program name="dummy_program_C2">
                <reference>dummy_model_01</reference>
            </program>
    
            <program name="dummy_program_C3">
                <reference>dummy_model_02</reference>
            </program>
    
            <program name="dummy_program_C4">
                <reference>dummy_model_03</reference>
                <identifiers>#,@</identifiers>
                <start_sequence>3</start_sequence>
                <required>true</required>
                <wait_exit>false</wait_exit>
                <expected_loading>5</expected_loading>
            </program>
        </programs>

    </application>

    <application name="dummy_application_D">
        <start_sequence>-1</start_sequence>
        <stop_sequence>100</stop_sequence>
        <starting_failure_strategy>CONTINUE</starting_failure_strategy>
        <running_failure_strategy>RESTART_APPLICATION</running_failure_strategy>

        <programs>
            <program pattern="dummies_">
                <reference>dummy_model_03</reference>
            </program>
    
            <program pattern="^dummies_01_">
                <identifiers>#</identifiers>
                <start_sequence>1</start_sequence>
                <stop_sequence>1</stop_sequence>
                <required>false</required>
                <wait_exit>true</wait_exit>
                <expected_loading>75</expected_loading>
            </program>

            <program pattern="d.*02.*">
                <reference>dummy_model_04</reference>
            </program>
        </programs>

    </application>

</root>
'''

# Contents of a rules file with no schema error
XmlTest = b'''\
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<root>
    <alias name="nodes_model_03">10.0.0.4, 10.0.0.2</alias>
    <alias name="nodes_appli_D">10.0.0.1, 10.0.0.5</alias>
    <alias name="not used">10.0.0.2, nodes_appli_D</alias>

    <model name="dummy_model_01">
        <stop_sequence>0</stop_sequence>
        <required>false</required>
        <wait_exit>false</wait_exit>
        <expected_loading>25</expected_loading>
        <running_failure_strategy>STOP_APPLICATION</running_failure_strategy>
    </model>

    <model name="dummy_model_02">
        <identifiers>#</identifiers>
        <start_sequence>1</start_sequence>
        <required>true</required>
        <wait_exit>true</wait_exit>
    </model>

    <model name="dummy_model_03">
        <identifiers>nodes_model_03</identifiers>
        <stop_sequence>100</stop_sequence>
        <expected_loading>10</expected_loading>
    </model>

    <model name="dummy_model_04">
        <reference>dummy_model_01</reference>
        <expected_loading>20</expected_loading>
    </model>

    <model name="dummy_model_05">
        <reference>dummy_model_04</reference>
        <expected_loading>15</expected_loading>
    </model>

    <application name="dummy_application_A"/>

    <application name="dummy_application_B">
        <distribution>SINGLE_NODE</distribution>
        <start_sequence>1</start_sequence>
        <stop_sequence>4</stop_sequence>
        <starting_strategy>CONFIG</starting_strategy>
        <starting_failure_strategy>STOP</starting_failure_strategy>
        <running_failure_strategy>RESTART_PROCESS</running_failure_strategy>
        <operational_status>'.*B1' and 'dummy_program_B2'</operational_status>

        <programs>
            <program name="dummy_program_B0"/>
    
            <program name="dummy_program_B1">
                <identifiers>@</identifiers>
                <start_sequence>3</start_sequence>
                <stop_sequence>50</stop_sequence>
                <required>true</required>
                <wait_exit>false</wait_exit>
                <expected_loading>5</expected_loading>
                <running_failure_strategy>CONTINUE</running_failure_strategy>
             </program>
    
            <program name="dummy_program_B2">
                <identifiers>10.0.0.3</identifiers>
                <required>true</required>
                <running_failure_strategy>RESTART_PROCESS</running_failure_strategy>
            </program>
    
            <program name="dummy_program_B3">
                <identifiers>*</identifiers>
                <required>false</required>
                <expected_loading>100</expected_loading>
                <running_failure_strategy>STOP_APPLICATION</running_failure_strategy>
            </program>
    
            <program name="dummy_program_B4">
                <identifiers>10.0.0.3, 10.0.0.1, 10.0.0.5</identifiers>
                <running_failure_strategy>RESTART_APPLICATION</running_failure_strategy>
            </program>
        </programs>

    </application>

    <application name="dummy_application_C">
        <distribution>ALL_INSTANCES</distribution>
        <start_sequence>20</start_sequence>
        <stop_sequence>0</stop_sequence>
        <starting_strategy>LOCAL</starting_strategy>
        <starting_failure_strategy>ABORT</starting_failure_strategy>
        <running_failure_strategy>STOP_APPLICATION</running_failure_strategy>

        <programs>
            <program name="dummy_program_C0">
                <reference></reference>
            </program>
    
            <program name="dummy_program_C1">
                <reference>unknown</reference>
            </program>
    
            <program name="dummy_program_C2">
                <reference>dummy_model_01</reference>
            </program>
    
            <program name="dummy_program_C3">
                <reference>dummy_model_02</reference>
            </program>
        </programs>

    </application>

    <application pattern="application_D">
        <distribution>SINGLE_INSTANCE</distribution>
        <identifiers>nodes_appli_D</identifiers>
        <start_sequence>-1</start_sequence>
        <stop_sequence>100</stop_sequence>
        <starting_strategy>LESS_LOADED</starting_strategy>
        <starting_failure_strategy>CONTINUE</starting_failure_strategy>
        <running_failure_strategy>SHUTDOWN</running_failure_strategy>

        <programs>
            <program pattern="dummies_">
                <reference>dummy_model_03</reference>
                <start_sequence>50</start_sequence>
            </program>
    
            <program pattern="^d.*s_01_">
                <identifiers>#, 10.0.0.1, 10.0.0.5</identifiers>
                <start_sequence>1</start_sequence>
                <stop_sequence>1</stop_sequence>
                <required>false</required>
                <wait_exit>true</wait_exit>
                <expected_loading>75</expected_loading>
            </program>

            <program pattern="dum+ies_02_">
                <reference>dummy_model_04</reference>
            </program>
        </programs>

    </application>

    <application name="dummy_application_E">
        <starting_strategy>MOST_LOADED</starting_strategy>
        <programs>
            <program name="dummy_program_E">
                <reference>dummy_model_05</reference>
            </program>
        </programs>
    </application>

</root>
'''
